#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""The sardana submodule. It contains specific part of sardana"""

__all__ = ["BaseSardanaElement", "BaseSardanaElementContainer",
           "Pool", "MacroServer", "Door", "Sardana", "SardanaManager",
           "PoolElementType", "ControllerClassInfo", "ControllerInfo",
           "ChannelView", "PlotType", "Normalization", "AcqTriggerType",
           "AcqMode"]

#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# T E M P O R A R Y   I M P L E M E N T A T I O N
#
# THIS IS USED FOR TEST PURPOSES ONLY. DO NOT USE IT OUTSIDE SARDANA TESTS
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

__docformat__ = 'restructuredtext'

import PyTango

from taurus.core.util.enumeration import Enumeration
from taurus.core.util.log import Logger
from taurus.core.util.singleton import Singleton
from taurus.core.util.containers import CaselessDict
from taurus.core.util.codecs import CodecFactory

PoolElementType = Enumeration("PoolElementType",
    ("0D", "1D", "2D", "Communication", "CounterTimer", "IORegister",
      "Motor", "PseudoCounter", "PseudoMotor"))

ChannelView = Enumeration("ChannelView",
    ("Channel", "Enabled", "Output", "PlotType", "PlotAxes", "Timer",
     "Monitor", "Trigger", "Conditioning", "Normalization", "NXPath",
     "Shape", "DataType",
     "Unknown"))

PlotType = Enumeration("PlotType", ("No", "Spectrum", "Image"))

Normalization = Enumeration("Normalization", ("No", "Avg", "Integ"))

#: an enumeration describing all possible acquisition trigger types
AcqTriggerType = Enumeration("AcqTriggerType", (\
    "Software",  # channel triggered by software - start and stop by software
    "Gate",  # channel triggered by HW - start and stop by external
    "Unknown"))

#: an enumeration describing all possible acquisition mode types
AcqMode = Enumeration("AcqMode", (\
    "Timer",
    "Monitor",
    "Unknown"))


class BaseSardanaElement(object):
    """Generic sardana element"""

    def __init__(self, *args, **kwargs):
        self._manager = kwargs.pop('manager')
        self.__dict__.update(kwargs)
        self._data = kwargs
        self._object = None

    def __repr__(self):
        return "{0}({1})".format(self.type, self.full_name)

    def __str__(self):
        return self.name

    def __getattr__(self, name):
        return getattr(self.getObj(), name)

    def __cmp__(self, elem):
        return cmp(self.name, elem.name)

    def getData(self):
        return self._data

    def getName(self):
        return self.name

    def getId(self):
        return self.full_name

    def getType(self):
        return self.getTypes()[0]

    def getTypes(self):
        elem_types = self.type
        if isinstance(elem_types, (str, unicode)):
            return [elem_types]
        return elem_types

    def serialize(self, *args, **kwargs):
        kwargs.update(self._data)
        return kwargs

    def str(self, *args, **kwargs):
        #TODO change and check which is the active protocol to serialize
        #acordingly
        return CodecFactory().encode(('json', self.serialize(*args, **kwargs)))

    def getObj(self):
        obj = self._object
        if obj is None:
            self._object = obj = self._manager.getObject(self)
        return obj



class BaseSardanaElementContainer:

    def __init__(self):
        # dict<str, dict> where key is the type and value is:
        #     dict<str, MacroServerElement> where key is the element full name
        #                                   and value is the Element object
        self._type_elems_dict = CaselessDict()

        # dict<str, container> where key is the interface and value is the set
        # of elements which implement that interface
        self._interfaces_dict = {}

    def addElement(self, elem):
        elem_type = elem.getType()
        elem_full_name = elem.full_name

        #update type_elems
        type_elems = self._type_elems_dict.get(elem_type)
        if type_elems is None:
            self._type_elems_dict[elem_type] = type_elems = CaselessDict()
        type_elems[elem_full_name] = elem

        # update interfaces
        for interface in elem.interfaces:
            interface_elems = self._interfaces_dict.get(interface)
            if interface_elems is None:
                self._interfaces_dict[interface] = interface_elems = CaselessDict()
            interface_elems[elem_full_name] = elem

    def removeElement(self, e):
        elem_type = e.getType()

        # update type_elems
        type_elems = self._type_elems_dict.get(elem_type)
        if type_elems:
            del type_elems[e.full_name]

        # update interfaces
        for interface in e.interfaces:
            interface_elems = self._interfaces_dict.get(interface)
            del interface_elems[e.full_name]

    def removeElementsOfType(self, t):
        for elem in self.getElementsOfType(t):
            self.removeElement(elem)

    def getElementsOfType(self, t):
        elems = self._type_elems_dict.get(t, {})
        return elems

    def getElementNamesOfType(self, t):
        return [e.name for e in self.getElementsOfType(t).values()]

    def getElementsWithInterface(self, interface):
        elems = self._interfaces_dict.get(interface, {})
        return elems

    def getElementsWithInterfaces(self, interfaces):
        ret = CaselessDict()
        for interface in interfaces:
            ret.update(self.getElementsWithInterface(interface))
        return ret

    def getElementNamesWithInterface(self, interface):
        return [e.name for e in self.getElementsWithInterface(interface).values()]

    def hasElementName(self, elem_name):
        return self.getElement(elem_name) != None

    def getElement(self, elem_name):
        elem_name = elem_name.lower()
        for elems in self._type_elems_dict.values():
            elem = elems.get(elem_name)  # full_name?
            if elem is not None:
                return elem
            for elem in elems.values():
                if elem.name.lower() == elem_name:
                    return elem

    def getElementWithInterface(self, elem_name, interface):
        elem_name = elem_name.lower()
        elems = self._interfaces_dict.get(interface, {})
        if elem_name in elems:
            return elems[elem_name]
        for elem in elems.values():
            if elem.name.lower() == elem_name:
                return elem

    def getElements(self):
        ret = set()
        for elems in self._type_elems_dict.values():
            ret.update(elems.values())
        return ret

    def getInterfaces(self):
        return self._interfaces_dict

    def getTypes(self):
        return self._type_elems_dict

#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# T E M P O R A R Y   I M P L E M E N T A T I O N
#
# THIS IS USED FOR TEST PURPOSES ONLY. DO NOT USE IT OUTSIDE SARDANA TESTS
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
class PropertyInfo():
    def __init__(self, name, type, format, default_value=None):
        self._name = name
        self._type = type
        self._format = format
        self._default_value = default_value

    def get_name(self):
        return self._name

    def get_type(self):
        return self._type

    def get_format(self):
        return self._format

    def get_default_value(self):
        return self._default_value


#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# T E M P O R A R Y   I M P L E M E N T A T I O N
#
# THIS IS USED FOR TEST PURPOSES ONLY. DO NOT USE IT OUTSIDE SARDANA TESTS
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
class ControllerClassInfo(object):

    def __init__(self, name, type, library):
        self._name = name
        self._type = type
        self._library = library

    def get_max_elements(self):
        return 20

    def get_name(self):
        return self._name

    def get_model(self):
        # fake data ###############
        return "Model of " + self._name

    def get_icon(self):
        # fake data ###############
        import taurus.qt.qtgui.resource

        return taurus.qt.qtgui.resource.getIcon(":/designer/extra_motor.png")

    def get_organization(self):
        # fake data ###############
        return "Organization of " + self._name

    def get_description(self):
        #fake data############
        descr = "This is description of "
        for i in range(20):
            descr = descr + " and " + self._name
        ####################
        return descr

    def get_family(self):
        # fake data ###############
        return "Family of " + self._name

    def get_properties(self):
        properties = []
        # fake data ######################
        properties.append(PropertyInfo("my parameter", "string", "0D", "deviceName"))
        properties.append(PropertyInfo("asdsadasd", "integer", "0D", 5))
        properties.append(PropertyInfo("boollll0", "boolean", "0D", False))
        properties.append(PropertyInfo("boollll0", "boolean", "0D", True))
        properties.append(PropertyInfo("boollll0", "boolean", "0D", False))
        properties.append(PropertyInfo("number1", "float", "0D", 3.5))
        properties.append(PropertyInfo("string2", "string", "0D", "hehe"))
        properties.append(PropertyInfo("tableIntegerD1", "integer", "1D", [1, 2, 3]))
        properties.append(PropertyInfo("tablestringD1", "string", "1D", ["aaaa", "bbb", "ccc"]))
        properties.append(PropertyInfo("tablefloatD1", "float", "1D", [1.0, 2.5, 3.6]))
        properties.append(PropertyInfo("tablebooleanD1", "boolean", "1D", [True, False, True, False]))
        properties.append(PropertyInfo("tablebleintegerD1", "integer", "1D", [1, 2, 3]))
        properties.append(PropertyInfo("tablebooleanD2", "boolean", "2D", [ [True, False, True], [True, True, True], [False, False, False] ]))
        properties.append(PropertyInfo("tableinteger2", "integer", "2D", [ [1, 2, 3], [11, 22, 33], [-10, -20, -30] ]))
        properties.append(PropertyInfo("tablefloatD2", "float", "2D", [ [0.5, 0.6, 0.8], [0.4, 0.0, 0.333333], [-0.1111, 1, 123123.6] ]))
        properties.append(PropertyInfo("tablestringD2", "string", "2D", [ ["aaaa", "bbb", "ccc"], ["aaaa2", "bbb2", "ccc2"], ["aaaa3", "bbb3", "ccc3"] ]))

        return properties

    def get_controller_type(self):
        return self._type


#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# T E M P O R A R Y   I M P L E M E N T A T I O N
#
# THIS IS USED FOR TEST PURPOSES ONLY. DO NOT USE IT OUTSIDE SARDANA TESTS
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
class ControllerInfo(object):

    def __init__(self, name, ctrl_class_info):
        self._name = name
        self._ctrl_class_info = ctrl_class_info

    def get_controller_class_info(self):
        return self._ctrl_class_info

    def get_controller_type(self):
        return self._ctrl_class_info.get_controller_type()

    def get_name(self):
        return self._name

    def get_max_elements(self):
        return self._ctrl_class_info.get_max_elements()

    def is_axis_free(self, axis):
        #fake data
        if axis == 3:
            return False
        else:
            return True

    def is_name_free(self, name):
        #fake data
        if name == "asd":
            return False
        else:
            return True

    def get_icon(self):
        return self._ctrl_class_info.get_icon()

#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# T E M P O R A R Y   I M P L E M E N T A T I O N
#
# THIS IS USED FOR TEST PURPOSES ONLY. DO NOT USE IT OUTSIDE SARDANA TESTS
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
class Pool(object):

    def __init__(self, sardana, name, poolpath, version, alias=None, device_name=None):
        self._sardana = sardana
        self._name = name
        self._poolpath = poolpath
        self._version = version
        self._alias = alias
        self._device_name = device_name

    def starter_run(self, host, level=1):
        return True

    def get_name(self):
        return self._name

    def local_run(self):
        return True

    def get_element_types(self):
        return sorted(PoolElementType.keys())

    def get_controller_class_infos(self):
        #fake data ########################
        data = []
        for i in range(5):
            data.append(ControllerClassInfo("motorController" + str(i), PoolElementType.Motor, None))
        for i in range(5):
            data.append(ControllerClassInfo("counterTimerController" + str(i), PoolElementType.CounterTimer, None))

        return data

    def get_controller_infos(self):
        ctrl_classes = self.get_controller_class_infos()
        data = []
        for i in range(2):
            data.append(ControllerInfo("My_motor_ctrl_" + str(i), ctrl_classes[i]))
        for i in range(2):
            data.append(ControllerInfo("My_ct_ctrl_" + str(i), ctrl_classes[i + 5]))
        return data

    def create_controller(self, controller_class_info, name, properties):
        pass

    def create_element(self, controller_name, name, axis):
        pass


#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# T E M P O R A R Y   I M P L E M E N T A T I O N
#
# THIS IS USED FOR TEST PURPOSES ONLY. DO NOT USE IT OUTSIDE SARDANA TESTS
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
class MacroServer(object):

    def __init__(self, sardana, name, macropath, pool_names, version, alias=None, device_name=None):
        self._sardana = sardana
        self._name = name
        self._macropath = macropath
        self._pool_names = pool_names
        self._version = version
        self._alias = alias
        self._device_name = device_name
        self._doors = []

    def create_door(self, alias, device_name):
        try:
            return self._create_door(alias, device_name)
        except:
            db = self.get_database()
            db.delete_device(device_name)
            raise

    def _create_door(self, alias, device_name):
        db = self.get_database()
        info = PyTango.DbDevInfo()
        info.name = device_name
        info._class = "Door"
        info.server = "MacroServer/" + self._name
        db.add_device(info)
        if alias:
            db.put_device_alias(device_name, alias)
        door = Door(alias=alias, device_name=device_name)
        self._doors.append(door)
        return door

    def remove_door(self, device_name):
        pass

    def starter_run(self, host, level=1):
        return True

    def local_run(self):
        return True

    def get_database(self):
        return self._sardana.get_database()

#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# T E M P O R A R Y   I M P L E M E N T A T I O N
#
# THIS IS USED FOR TEST PURPOSES ONLY. DO NOT USE IT OUTSIDE SARDANA TESTS
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
class Door(object):

    def __init__(self, alias=None, device_name=None):
        self._name = alias
        self._device_name = device_name


#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# T E M P O R A R Y   I M P L E M E N T A T I O N
#
# THIS IS USED FOR TEST PURPOSES ONLY. DO NOT USE IT OUTSIDE SARDANA TESTS
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
class Sardana(object):

    def __init__(self, sardana_db , name, device_name=None):
        self._sardana_db = sardana_db
        self._name = name
        self._device_name = device_name
        self._pools = []
        self._macroservers = []
        self._init()

    def _init(self):
        if not self._device_name:
            return
        self._pools = []
        self._macroservers = []
        dev_name = self._device_name
        db = self.get_database()
        cache = db.cache()
        dev_info = cache.devices()[dev_name]
        dev_class_name = dev_info.klass().name()
        if dev_class_name == "Pool":
            pass
        elif dev_class_name == "MacroServer":
            ms_dev_name = dev_name
            ms_prop_list = map(str.lower, db.get_device_property_list(ms_dev_name, "*"))
            ms_props = db.get_device_property(ms_dev_name, ms_prop_list)
            ms_name = dev_info.server().serverInstance()
            ms_alias = dev_info.alias()
            ms = MacroServer(self, ms_name, ms_props.get("macropath"), ms_props.get("poolnames"),
                             ms_props.get("version"), ms_alias, ms_dev_name)
            self._macroservers.append(ms)
            for pool_dev_name in ms_props.get("poolnames", ()):
                pool_prop_list = map(str.lower, db.get_device_property_list(pool_dev_name, "*"))
                pool_props = db.get_device_property(pool_dev_name, pool_prop_list)
                pool_dev_info = cache.devices()[pool_dev_name]
                pool_name = pool_dev_info.server().serverInstance()
                pool_alias = pool_dev_info.alias()
                pool = Pool(self, pool_name, pool_props.get("poolpath"), pool_props.get("version"), pool_alias, pool_dev_name)
                self._pools.append(pool)

    def get_name(self):
        return self._name

    def set_device_name(self, device_name):
        self._device_name = device_name
        self._init()

    def get_device_name(self):
        return self._device_name

    def get_pools(self):
        return self._pools

    def get_macro_servers(self):
        return self._macro_servers

    def create_pool(self, name, poolpath, version, alias=None, device_name=None):
        try:
            return self._create_pool(name, poolpath, version, alias=alias, device_name=device_name)
        except:
            db = self.get_database()
            db.delete_device(device_name)
            raise

    def _create_pool(self, name, poolpath, version, alias=None, device_name=None):
        db = self.get_database()
        info = PyTango.DbDevInfo()
        info.name = device_name
        info._class = "Pool"
        info.server = "Pool/" + name
        db.add_device(info)
        if alias:
            db.put_device_alias(device_name, alias)

        db.put_device_property(device_name, {"PoolPath" : poolpath, "Version": version})
        pool = Pool(self, name, poolpath, version, alias=alias, device_name=device_name)
        self._pools.append(pool)
        db.cache().refresh()
        return pool

    def create_macroserver(self, name, macropath, pool_names, version, alias=None, device_name=None):
        try:
            return self._create_macroserver(name, macropath, pool_names, version, alias=alias, device_name=device_name)
        except:
            db = self.get_database()
            db.delete_device(device_name)
            raise

    def _create_macroserver(self, name, macropath, pool_names, version, alias=None, device_name=None):
        db = self.get_database()
        info = PyTango.DbDevInfo()
        info.name = device_name
        info._class = "MacroServer"
        info.server = "MacroServer/" + name
        db.add_device(info)
        if alias:
            db.put_device_alias(device_name, alias)

        db.put_device_property(device_name, {"MacroPath" : macropath, "Version": version, "PoolNames":pool_names})
        ms = MacroServer(self, name, macropath, pool_names, version, alias=alias, device_name=device_name)
        self._macroservers.append(ms)
        db.cache().refresh()
        return ms

    def remove_pool(self):
        pass

    def remove_macroserver(self):
        pass

    def get_database(self):
        return self._sardana_db.get_database()

#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# T E M P O R A R Y   I M P L E M E N T A T I O N
#
# THIS IS USED FOR TEST PURPOSES ONLY. DO NOT USE IT OUTSIDE SARDANA TESTS
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
class DatabaseSardana(object):
    """A class containning all sardanas for a single database"""

    def __init__(self, db):
        assert(db is not None)
        self._db = db
        self.refresh()

    def refresh(self):
        self._sardanas = sardanas = {}
        services = self._db.get_service_list("Sardana/.*")
        for service, dev in services.items():
            service_type, service_instance = service.split("/", 1)
            try:
                sardanas[service_instance] = Sardana(self, service_instance, dev)
            except:
                pass

    def create_sardana(self, name, device_name):
        if self._sardanas.has_key(name):
            raise Exception("Sardana '%s' already exists" % name)
        self._db.register_service("Sardana", name, device_name)
        sardana = Sardana(self, name)
        self._sardanas[name] = sardana
        return sardana

    def remove_sardana(self, name):
        try:
            self._sardanas.pop(name)
        except KeyError:
            raise Exception("Sardana '%s' does NOT exist" % name)
        self._db.unregister_service("Sardana", name)

    def get_sardanas(self):
        return self._sardanas

    def get_sardana(self, name):
        return self._sardanas[name]

    def get_database(self):
        return self._db

#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# T E M P O R A R Y   I M P L E M E N T A T I O N
#
# THIS IS USED FOR TEST PURPOSES ONLY. DO NOT USE IT OUTSIDE SARDANA TESTS
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
class SardanaManager(Singleton, Logger):

    def __init__(self):
        """ Initialization. Nothing to be done here for now."""
        pass

    def init(self, *args, **kwargs):
        """Singleton instance initialization.
           **For internal usage only**"""
        name = self.__class__.__name__
        self.call__init__(Logger, name)
        self._db_sardanas = {}

    def _get_db_sardana(self, db=None):
        if db is None:
            import taurus
            db = taurus.Database()
        db_sardana = self._db_sardanas.get(db)
        if db_sardana is None:
            db_sardana = DatabaseSardana(db)
        return db_sardana

    def create_sardana(self, name, device_name, db=None):
        return self._get_db_sardana(db).create_sardana(name, device_name)

    def remove_sardana(self, name, db=None):
        self._get_db_sardana(db).remove_sardana(name)

    def get_sardanas(self, db=None):
        return self._get_db_sardana(db).get_sardanas()

    def get_sardana(self, name, db=None):
        return self._get_db_sardana(db).get_sardana(name)

    def get_hosts(self):
        return ["localhost"] + ["controls%02d" % i for i in range(5)]

    def get_level_range(self):
        return 1, 200

    def has_localhost_starter(self):
        import socket
        return socket.gethostname() in self.get_hosts()

    @classmethod
    def get_default_pool_path(cls):
        pathList = []
        pathList.append("/homelocal/sicilia/lib/poolcontrollers")
        pathList.append("/homelocal/sicilia/lib/python/site-packages/poolcontrollers")
        return pathList

    @classmethod
    def get_default_ms_path(cls):
        pathList = []
        pathList.append("/homelocal/sicilia/lib/python/site-packages/macroserver/macros")
        return pathList








#!/usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This package provides the spock generic utilities"""

__all__ = ['page', 'arg_split', 'get_gui_mode', 'get_pylab_mode',
           'get_color_mode', 'get_app',
           'get_shell', 'get_ipapi', 'get_config', 'get_editor', 'ask_yes_no',
           'spock_input',
           'translate_version_str2int', 'get_ipython_version',
           'get_ipython_version_number', 'get_python_version',
           'get_python_version_number', 'get_ipython_dir',
           'get_ipython_profiles',
           'get_pytango_version', 'get_pytango_version_number',
           'get_server_for_device', 'get_macroserver_for_door',
           'get_device_from_user', 'get_tango_db', 'get_tango_host_from_user',
           'print_dev_from_class', 'from_name_to_tango', 'clean_up',
           'get_taurus_core_version', 'get_taurus_core_version_number',
           'check_requirements', 'get_door', 'get_macro_server',
           'expose_magic', 'unexpose_magic', 'expose_variable',
           'expose_variables', 'unexpose_variable',
           'create_spock_profile', 'check_for_upgrade', 'get_args',
           'start', 'mainloop', 'run',
           'load_ipython_extension', 'unload_ipython_extension', 'load_config',
           'MSG_FAILED', 'MSG_FAILED_WR', 'MSG_R', 'MSG_ERROR',
           'MSG_DONE', 'MSG_OK']

__docformat__ = 'restructuredtext'

import sys
import os
import socket

import IPython
import IPython.core.magic
from IPython.core.page import page
from IPython.core.profiledir import ProfileDirError, ProfileDir
from IPython.core.application import BaseIPythonApplication
from IPython.core.interactiveshell import InteractiveShell
from IPython.utils.io import ask_yes_no as _ask_yes_no
from IPython.utils.path import get_ipython_dir
from IPython.utils.process import arg_split
from IPython.utils.coloransi import TermColors
from IPython.config.application import Application
from IPython.terminal.ipapp import TerminalIPythonApp, launch_new_instance

import taurus
#from taurus.core import Release as TCRelease

from taurus.core.taurushelper import Factory
from taurus.core.util.codecs import CodecFactory

# make sure Qt is properly initialized
from taurus.external.qt import Qt

from sardana.spock import exception
from sardana.spock import colors
from sardana.spock import release

SpockTermColors = colors.TermColors

requirements = {
#     module     minimum  recommended
    "IPython"     : ("0.11.0", "0.12.0"),
    "Python"      : ("2.6.0", "2.6.0"),
    "PyTango"     : ("7.2.0", "7.2.3"),
    "taurus.core" : ("3.0.0", "3.0.0")
}

ENV_NAME = "_E"

#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# IPython utilities
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

def get_gui_mode():
    return 'qt'

def get_pylab_mode():
    return get_app().pylab

def get_color_mode():
    return get_config().InteractiveShell.colors

def get_app():
    #return TerminalIPythonApp.instance()
    return Application.instance()

def get_shell():
    """Get the global InteractiveShell instance."""
    return get_app().shell

def get_ipapi():
    """Get the global InteractiveShell instance."""
    return InteractiveShell.instance()

def get_config():
    return get_app().config

def get_editor():
    return get_ipapi().editor

def ask_yes_no(prompt,default=None):
    """Asks a question and returns a boolean (y/n) answer.

    If default is given (one of 'y','n'), it is used if the user input is
    empty. Otherwise the question is repeated until an answer is given.

    An EOF is treated as the default answer.  If there is no default, an
    exception is raised to prevent infinite loops.

    Valid answers are: y/yes/n/no (match is not case sensitive)."""

    if default:
        prompt = '%s [%s]' % (prompt, default)
    return _ask_yes_no(prompt, default)

def spock_input(prompt='',  ps2='... '):
    return raw_input(prompt)

def translate_version_str2int(version_str):
    """Translates a version string in format x[.y[.z[...]]] into a 000000 number"""
    import math
    parts = version_str.split('.')
    i, v, l = 0, 0, len(parts)
    if not l: return v
    while i<3:
        try:
            v += int(parts[i])*int(math.pow(10,(2-i)*2))
            l -= 1
            i += 1
        except ValueError:
            return v
        if not l: return v
    return v

    try:
        v += 10000*int(parts[0])
        l -= 1
    except ValueError:
        return v
    if not l: return v

    try:
        v += 100*int(parts[1])
        l -= 1
    except ValueError:
        return v
    if not l: return v

    try:
        v += int(parts[0])
        l -= 1
    except ValueError:
        return v
    if not l: return v

def get_ipython_version():
    """Returns the current IPython version"""
    v = None
    try:
        try:
            v = IPython.Release.version
        except Exception:
            try:
                v = IPython.release.version
            except Exception, e2:
                print e2
    except Exception, e3:
        print e3
    return v

def get_ipython_version_number():
    """Returns the current IPython version number"""
    ipyver_str = get_ipython_version()
    if ipyver_str is None: return None
    return translate_version_str2int(ipyver_str)

def get_python_version():
    return '.'.join(map(str,sys.version_info[:3]))

def get_python_version_number():
    pyver_str = get_python_version()
    return translate_version_str2int(pyver_str)

def get_ipython_profiles(path=None):
    """list profiles in a given root directory"""
    if path is None:
        path = get_ipython_dir()
    files = os.listdir(path)
    profiles = []
    for f in files:
        full_path = os.path.join(path, f)
        if os.path.isdir(full_path) and f.startswith('profile_'):
            profiles.append(f.split('_',1)[-1])
    return profiles

#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# PyTango utilities
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

def get_pytango_version():
    try:
        import PyTango
        try:
            return PyTango.Release.version
        except:
            return '0.0.0'
    except:
        return None

def get_pytango_version_number():
    tgver_str = get_pytango_version()
    if tgver_str is None: return None
    return translate_version_str2int(tgver_str)

def get_server_for_device(device_name):
    db = get_tango_db()
    device_name = device_name.lower()
    server_list = db.get_server_list()
    for server in server_list:
        for dev in db.get_device_class_list(server)[::2]:
            if dev.lower() == device_name:
                return server
    return None

def get_macroserver_for_door(door_name):
    """Returns the MacroServer device name in the same DeviceServer as the
    given door device"""
    _, door_name, _ = from_name_to_tango(door_name)
    db = get_tango_db()
    door_name = door_name.lower()
    server_list = list(db.get_server_list('MacroServer/*'))
    server_list += list(db.get_server_list('Sardana/*'))
    server_devs = None
    for server in server_list:
        server_devs = db.get_device_class_list(server)
        devs, klasses = server_devs[0::2], server_devs[1::2]
        for dev in devs:
            if dev.lower() == door_name:
                for i, klass in enumerate(klasses):
                    if klass == 'MacroServer':
                        return "%s:%s/%s" % (db.get_db_host(), db.get_db_port(), devs[i])
    else:
        return None

def get_device_from_user(expected_class, dft = None):
    """Gets a device of the given device class from user input"""
    dft = print_dev_from_class(expected_class, dft)
    prompt = "%s name from the list" % expected_class
    if not dft is None:
        prompt += "[%s]" % dft
    prompt += "? "
    from_user = raw_input(prompt).strip() or dft

    name = ''
    try:
        full_name, name, _ = from_name_to_tango(from_user)
    except:
        print "Warning: the given %s does not exist" % expected_class
        return name

    try:
        db = get_tango_db()
        cl_name = db.get_class_for_device(name)
        class_correct = cl_name == expected_class
        if not class_correct:
            print "Warning: the given name is not a %s (it is a %s)"%(expected_class,cl_name)
    except Exception as e:
        print "Warning: unable to confirm if '%s' is valid" % name
        print str(e)
    return full_name

def get_tango_db():
    import PyTango
    tg_host = PyTango.ApiUtil.get_env_var("TANGO_HOST")

    db = None
    if tg_host is None:
        host,port = get_tango_host_from_user()
        tg_host = "%s:%d" % (host,port)
        os.environ["TANGO_HOST"] = tg_host
        db = PyTango.Database()
    else:
        try:
            db = PyTango.Database()
        except:
            # tg host is not valid. Find a valid one
            host,port = get_tango_host_from_user()
            tg_host = "%s:%d" % (host,port)
            os.environ["TANGO_HOST"] = tg_host

            db = PyTango.Database()
    return db

def get_tango_host_from_user():
    import PyTango
    while True:
        prompt = "Please enter a valid tango host (<host>:<port>): "
        from_user = raw_input(prompt).strip()

        try:
            host, port = from_user.split(':')
            try:
                port = int(port)
                try:
                    socket.gethostbyname(host)
                    try:
                        PyTango.Database(host, port)
                        return (host, port)
                    except:
                        exp = "No tango database found at %s:%d" % (host, port)
                except:
                    exp = "Invalid host name %s" % host
            except:
                exp = "Port must be a number > 0"
        except:
            exp = "Invalid tango host. Must be in format <host>:<port>"
        exp = "Invalid tango host. %s " % exp
        print exp

def print_dev_from_class(classname, dft = None):

    db = get_tango_db()
    pytg_ver = get_pytango_version_number()
    if pytg_ver >= 030004:
        server_wildcard = '*'
        try:
            exp_dev_list = db.get_device_exported_for_class(classname)
        except:
            exp_dev_list = []
    else:
        server_wildcard = '%'
        exp_dev_list = []

    res = None
    dev_list = db.get_device_name(server_wildcard,classname)
    tg_host = "%s:%s" % (db.get_db_host(),db.get_db_port())
    print "Available",classname,"devices from",tg_host,":"
    for dev in dev_list:
        _, name, alias = from_name_to_tango(dev)
        out = alias or name
        if alias: out += ' (a.k.a. %s)' % name
        out = "%-25s" % out
        if dev in exp_dev_list:
            out += " (running)"
        print out
        if dft:
            if dft.lower() == name.lower(): res = name
            elif not alias is None and dft.lower() == alias.lower(): res = alias

    return res

def from_name_to_tango(name):

    db = get_tango_db()

    alias = None

    c = name.count('/')
    # if the db prefix is there, remove it first
    if c == 3 or c == 1:
        name = name[name.index("/")+1:]

    elems = name.split('/')
    l = len(elems)

    if l == 3:
        try:
            alias = db.get_alias(name)
            if alias.lower() == 'nada':
                alias = None
        except:
            alias = None
    elif l == 1:
        alias = name
        name = db.get_device_alias(alias)
    else:
        raise Exception("Invalid device name '%s'" % name)

    full_name = "%s:%s/%s" % (db.get_db_host(), db.get_db_port(), name)
    return full_name, name, alias

#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# taurus utilities
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

def clean_up():
    taurus.Manager().cleanUp()

def get_taurus_core_version():
    try:
        import taurus
        return taurus.core.release.version
    except:
        import traceback
        traceback.print_exc()
        return '0.0.0'

def get_taurus_core_version_number():
    tgver_str = get_taurus_core_version()
    if tgver_str is None: return None
    return translate_version_str2int(tgver_str)

#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# Requirements checking
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

def check_requirements():
    r = requirements
    minPyTango, recPyTango = map(translate_version_str2int, r["PyTango"])
    minIPython, recIPython = map(translate_version_str2int, r["IPython"])
    minPython, recPython = map(translate_version_str2int, r["Python"])
    minTaurusCore, recTaurusCore = map(translate_version_str2int, r["taurus.core"])

    currPython = get_python_version_number()
    currIPython = get_ipython_version_number()
    currPyTango = get_pytango_version_number()
    currTaurusCore = get_taurus_core_version_number()

    errMsg = ""
    warnMsg = ""

    errPython, errIPython, errPyTango, errTaurusCore = False, False, False, False
    if currPython is None:
        errMsg += "Spock needs Python version >= %s. No python installation found\n" % requirements["Python"][0]
        errPython = True
    elif currPython < minPython:
        errMsg += "Spock needs Python version >= %s. Current version is %s\n" % (requirements["Python"][0], get_python_version())
        errPython = True

    if currIPython is None:
        errMsg += "Spock needs IPython version >= %s. No IPython installation found\n" % requirements["IPython"][0]
        errIPython = True
    elif currIPython < minIPython:
        errMsg += "Spock needs IPython version >= %s. Current version is %s\n" % (requirements["IPython"][0], get_ipython_version())
        errIPython = True

    if currPyTango is None:
        errMsg += "Spock needs PyTango version >= %s. No PyTango installation found\n" % requirements["IPython"][0]
        errPyTango = True
    elif currPyTango < minPyTango:
        errMsg += "Spock needs PyTango version >= %s. " % requirements["PyTango"][0]
        if currPyTango > 0:
            errMsg += "Current version is %s\n" % get_pytango_version()
        else:
            errMsg += "Current version is unknown (most surely too old)\n"
        errPyTango = True

    if currTaurusCore is None:
        errMsg += "Spock needs taurus.core version >= %s. No taurus.core installation found\n" % requirements["taurus.core"][0]
        errTaurusCore = True
    elif currTaurusCore < minTaurusCore:
        errMsg += "Spock needs taurus.core version >= %s. " % requirements["taurus.core"][0]
        if currTaurusCore > 0:
            errMsg += "Current version is %s\n" % get_taurus_core_version()
        else:
            errMsg += "Current version is unknown (most surely too old)\n"
        errTaurusCore = True

    # Warnings
    if not errPython and currPython < recPython:
        warnMsg += "Spock recommends Python version >= %s. Current version " \
                   "is %s\n" % (requirements["Python"][1],
                                get_python_version())

    if not errIPython and currIPython < recIPython:
        warnMsg += "Spock recommends IPython version >= %s. Current version " \
                   "is %s\n" % (requirements["IPython"][1],
                                get_ipython_version())

    if not errPyTango and currPyTango < recPyTango:
        warnMsg += "Spock recommends PyTango version >= %s. Current version " \
                   "is %s\n" % (requirements["PyTango"][1],
                                get_pytango_version())

    if not errTaurusCore and currTaurusCore < recTaurusCore:
        warnMsg += "Spock recommends taurus.core version >= %s. Current " \
                   "version is %s\n" % (requirements["taurus.core"][1],
                                        get_taurus_core_version())

    if errMsg:
        errMsg += warnMsg
        raise exception.SpockMissingRequirement, errMsg

    if warnMsg:
        raise exception.SpockMissingRecommended, warnMsg

    return True

def _get_dev(dev_type):
    spock_config = get_config().Spock
    taurus_dev = None
    taurus_dev_var = "_" + dev_type
    if hasattr(spock_config, taurus_dev_var):
        taurus_dev = getattr(spock_config, taurus_dev_var)
    if taurus_dev is None:
        dev_name = getattr(spock_config, dev_type + '_name')
        factory = Factory()
        taurus_dev = factory.getDevice(dev_name)
        import PyTango
        dev = PyTango.DeviceProxy(dev_name)
        setattr(spock_config, dev_type, dev)
        setattr(spock_config, taurus_dev_var, taurus_dev)
        shell = get_shell()
        dev_type_upper = dev_type.upper()
        shell.user_ns[dev_type_upper] = dev
        shell.user_ns["_" + dev_type_upper] = taurus_dev
    return taurus_dev

def get_door():
    return _get_dev('door')

def get_macro_server():
    return _get_dev('macro_server')

def _macro_completer(self, event):
    """Method called by the IPython autocompleter. It will determine possible
       values for macro arguments.
    """
    ms = get_macro_server()

    macro_name = event.command.lstrip('%')
    # calculate parameter index
    param_idx = len(event.line.split()) - 1
    if not event.line.endswith(' '): param_idx -= 1
    # get macro info
    info = ms.getMacroInfoObj(macro_name)
    # if macro doesn't have parameters return
    if param_idx < 0 or not info.hasParams() : return
    # get the parameter info
    possible_params = info.getPossibleParams(param_idx)
    # return the existing elements for the given parameter type
    if possible_params:
        res = []
        for param in possible_params:
            res.extend(ms.getElementNamesWithInterface(param['type']))
        return res

def expose_magic(name, fn, completer_func=_macro_completer):
    shell = get_shell()
    fn.old_magic = shell.define_magic(name, fn)
    fn.old_completer = completer_func

    if completer_func is None:
        return

    # enable macro param completion
    if completer_func is not None:
        shell.set_hook('complete_command', completer_func, str_key = name)
        shell.set_hook('complete_command', completer_func, str_key = '%'+name)

def unexpose_magic(name):
    shell = get_shell()
    mg_name = 'magic_' + name
    if hasattr(shell, mg_name):
        magic_fn  = getattr(shell, mg_name)
        delattr(shell, mg_name)
        if hasattr(magic_fn, 'old_magic') and magic_fn.old_magic is not None:
            expose_magic(name, magic_fn.old_magic, magic_fn.old_completer)

def expose_variable(name, value):
    get_shell().user_ns[name] = value

def expose_variables(d):
    get_shell().user_ns.update(d)

def unexpose_variable(name):
    user_ns = get_shell().user_ns
    del user_ns[name]

def create_spock_profile(userdir, dft_profile, profile, door_name=None):
    """Create a profile file from a profile template file """
    if not os.path.isdir(userdir):
        ProfileDir.create_profile_dir(userdir)
    p_dir = ProfileDir.create_profile_dir_by_name(userdir, profile)
    ###########################################################################
    # NOTE: BaseIPythonApplication.config_file_name.default_value should return
    # the config file name, but it returns an empty string instead (at least 
    # in some cases). For now, we give a hardcoded name if it is empty
    # TODO: Check why this is the case
    config_file_name = BaseIPythonApplication.config_file_name.default_value 
    config_file_name = config_file_name or 'ipython_config.py' 
    ###########################################################################
    abs_config_file_name = os.path.join(p_dir.location, config_file_name)
    create_config = True
    if os.path.isfile(abs_config_file_name):
        create_config = ask_yes_no("Spock configuration file already exists. "\
                                   "Do you wish to replace it?", default='y')
    if not create_config:
        return

    src_data = """\
\"\"\"Settings for Spock session\"\"\"

#
# Please do not delete the next lines has they are used to check the version
# number for possible upgrades
# spock_creation_version = {version}
# door_name = {door_name}
#

import PyTango.ipython
import sardana.spock.genutils
from sardana.spock.config import Spock

config = get_config()
config.Spock.macro_server_name = '{macroserver_name}'
config.Spock.door_name = '{door_name}'

load_subconfig('ipython_config.py', profile='default')
sardana.spock.load_config(config)

# Put any additional environment here and/or overwrite default sardana config
config.IPKernelApp.pylab = 'inline'

"""
    #
    # Discover door name
    #
    if door_name is None:
        door_name = get_device_from_user("Door", profile)
    else:
        full_door_name, door_name, _ = from_name_to_tango(door_name)
        door_name = full_door_name

    #
    # Discover macro server name
    #
    ms_name = get_macroserver_for_door(door_name)

    dest_data = src_data.format(version=release.version,
                                macroserver_name=ms_name,
                                door_name=door_name)

    sys.stdout.write('Storing %s in %s... ' % (config_file_name, p_dir.location))
    sys.stdout.flush()

    with file(abs_config_file_name, "w") as f:
        f.write(dest_data)
        f.close()
    sys.stdout.write(MSG_DONE + '\n')

def check_for_upgrade(ipy_profile_file, ipythondir, session, profile):
    # Check if the current profile is up to date with the spock version
    spock_profile_ver_str = '0.0.0'
    door_name = None

    # search for version and door inside the ipy_profile file
    for i, line in enumerate(ipy_profile_file):
        if i > 20 : break; # give up after 20 lines
        if line.startswith('# spock_creation_version = '):
            spock_profile_ver_str = line[line.index('=')+1:].strip()
        if line.startswith('# door_name = '):
            door_name = line[line.index('=')+1:].strip()

    # convert version from string to numbers
    spocklib_ver = translate_version_str2int(release.version)
    spock_profile_ver = translate_version_str2int(spock_profile_ver_str)

    if spocklib_ver == spock_profile_ver:
        return
    if spocklib_ver < spock_profile_ver:
        print '%sYour spock profile (%s) is newer than your spock version ' \
              '(%s)!' % (SpockTermColors.Brown, spock_profile_ver_str, release.version)
        print 'Please upgrade spock or delete the current profile %s' % SpockTermColors.Normal
        sys.exit(1)

    # there was no version track of spock profiles since spock 0.2.0 so change
    # the message
    if spock_profile_ver_str == '0.0.0':
        spock_profile_ver_str = '<= 0.2.0'
    msg = 'Your current spock door extension profile has been created with spock %s.\n' \
          'Your current spock door extension version is %s, therefore a profile upgrade is needed.\n' \
          % (spock_profile_ver_str, release.version)
    print msg
    prompt = 'Do you wish to upgrade now (warn: this will shutdown the current spock session) ([y]/n)? '
    r = raw_input(prompt) or 'y'
    if r.lower() == 'y':
        create_spock_profile(ipythondir, session, profile, door_name)
        sys.exit(0)

def get_args(argv):

    script_name = argv[0]
    _, session = os.path.split(script_name)
    script_name = os.path.realpath(script_name)

    macro_server = None
    door = None

    # Define the profile file
    profile = "spockdoor"
    try:
        for _, arg in enumerate(argv[1:]):
            if arg.startswith('--profile='):
                profile=arg[10:]
                break
        else:
            argv.append("--profile=" + profile)
    except:
        pass

    ipython_dir = get_ipython_dir()
    try:
        ProfileDir.find_profile_dir_by_name(ipython_dir, profile)
    except ProfileDirError:
        r = ''
        while not r in ('y','n'):
            prompt = 'Profile \'%s\' does not exist. Do you want to create '\
                     'one now ([y]/n)? ' % profile
            r = raw_input(prompt) or 'y'
        if r.lower() == 'y':
            create_spock_profile(ipython_dir, session, profile)
        else:
            sys.stdout.write('No spock door extension profile was created. Starting normal spock...\n')
            sys.stdout.flush()
            profile = ''

    # inform the shell of the profile it should use
    if not '--profile=' in argv and profile:
        argv.append('--profile=' + profile)

    user_ns = { 'MACRO_SERVER_NAME' : macro_server,
                'DOOR_NAME'         : door,
                'PROFILE'           : profile }

    return user_ns

#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# Useful constants
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

MSG_G = '[%s%%s%s]' % (SpockTermColors.Green, SpockTermColors.Normal)
MSG_R = '[%s%%s%s]' % (SpockTermColors.Red, SpockTermColors.Normal)
MSG_FAILED = MSG_R % 'FAILED'
MSG_FAILED_WR = MSG_R % 'FAILED: %s'
MSG_ERROR  = MSG_R % 'ERROR'
MSG_DONE   = MSG_G % 'DONE'
MSG_OK     = MSG_G % 'OK'

#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
# initialization methods
#-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

def init_taurus():
    # the CodecFactory is not thread safe. There are two attributes who will
    # request for it in the first event at startup in different threads
    # therefore this small hack: make sure CodecFactory is initialized.
    CodecFactory()

    factory = Factory()

    import sardana.spock.spockms
    macroserver = sardana.spock.spockms

    factory.registerDeviceClass('MacroServer', macroserver.SpockMacroServer)

    mode = get_gui_mode()
    if mode == 'qt':
        factory.registerDeviceClass('Door', macroserver.QSpockDoor)
    else:
        factory.registerDeviceClass('Door', macroserver.SpockDoor)


def load_ipython_extension(ipython):
    import sardana.spock.magic
    magic = sardana.spock.magic

    init_taurus()

    config = ipython.config
    user_ns = ipython.user_ns
    user_ns['MACRO_SERVER_NAME'] = config.Spock.macro_server_name
    user_ns['MACRO_SERVER_ALIAS'] = config.Spock.macro_server_alias
    user_ns['DOOR_NAME'] = config.Spock.door_name
    user_ns['DOOR_ALIAS'] = config.Spock.door_alias
    user_ns['DOOR_STATE'] = ""

    #shell.set_hook('late_startup_hook', magic.spock_late_startup_hook)
    ipython.set_hook('pre_prompt_hook', magic.spock_pre_prompt_hook)

    #if ip.IP.alias_table.has_key('mv'):
    #    del ip.IP.alias_table['mv']

    door = get_door()
    macro_server = get_macro_server()

    # Initialize the environment
    expose_variable(ENV_NAME, macro_server.getEnvironment())
    
    new_style_magics = hasattr(IPython.core.magic, 'Magics') and hasattr(IPython.core.magic, 'magics_class')
    
    if new_style_magics:
        @IPython.core.magic.magics_class
        class Sardana(IPython.core.magic.Magics):
            debug = IPython.core.magic.line_magic(magic.debug)
            www = IPython.core.magic.line_magic(magic.www)
            post_mortem = IPython.core.magic.line_magic(magic.post_mortem)
            spsplot = IPython.core.magic.line_magic(magic.post_mortem)
            macrodata = IPython.core.magic.line_magic(magic.macrodata)
            edmac = IPython.core.magic.line_magic(magic.edmac)
            showscan = IPython.core.magic.line_magic(magic.showscan)
            expconf = IPython.core.magic.line_magic(magic.expconf)
        
        ipython.register_magics(Sardana)
    else:
        expose_magic('debug', magic.debug, magic.debug_completer)
        expose_magic('www', magic.www, magic.www_completer)
        expose_magic('post_mortem', magic.post_mortem, magic.post_mortem_completer)
        expose_magic('spsplot', magic.spsplot, None)
        expose_magic('macrodata', magic.macrodata, None)
        expose_magic('edmac', magic.edmac, None)
        expose_magic('showscan', magic.showscan, None)
        expose_magic('expconf', magic.expconf, None)

    door.setConsoleReady(True)

def unload_ipython_extension(ipython):
    pass

def load_config(config):
    spockver = release.version
    pyver = get_python_version()
    ipyver = get_ipython_version()
    pytangover = get_pytango_version()
    tauruscorever = get_taurus_core_version()

    door = config.Spock.door_name

    if not hasattr(config.Spock, 'macro_server_name'):
        macro_server = get_macroserver_for_door(door)
    else:
        macro_server = config.Spock.macro_server_name

    full_door_tg_name, door_tg_name, door_tg_alias = from_name_to_tango(door)
    door_alias = door_tg_alias or door_tg_name
    full_ms_tg_name, ms_tg_name, ms_tg_alias = from_name_to_tango(macro_server)
    ms_alias = ms_tg_alias or ms_tg_name

    config.Spock.macro_server_name = full_ms_tg_name
    config.Spock.door_name = full_door_tg_name
    config.Spock.macro_server_alias = ms_alias
    config.Spock.door_alias = door_alias

    d = { "version" : spockver,
          "pyver" : pyver,
          "ipyver" : ipyver,
          "pytangover" : pytangover,
          "taurusver" : tauruscorever,
          #"profile" : ip.user_ns["PROFILE"],
          "door" : door_alias }

    d.update(TermColors.__dict__)

    gui_mode = get_gui_mode()

    banner = """\
%(Purple)sSpock %(version)s%(Normal)s -- An interactive laboratory application.

help      -> Spock's help system.
object?   -> Details about 'object'. ?object also works, ?? prints more.
"""
    banner = banner % d
    banner = banner.format(**d)

    ipy_ver = get_ipython_version_number()

    # ------------------------------------
    # Application
    # ------------------------------------
    app = config.Application
    app.log_level = 30

    # ------------------------------------
    # BaseIPythonApplication
    # ------------------------------------
    i_app = config.BaseIPythonApplication
    extensions = getattr(i_app, 'extensions', [])
    extensions.extend(["PyTango.ipython", "sardana.spock"])
    i_app.extensions = extensions

    # ------------------------------------
    # InteractiveShell
    # (IPython.core.interactiveshell)
    # ------------------------------------
    i_shell = config.InteractiveShell
    i_shell.autocall = 0
    i_shell.automagic = True
    i_shell.color_info = True
    i_shell.colors = 'Linux'
    i_shell.deep_reload = True
    i_shell.confirm_exit = False

    if ipy_ver >= 1200:
        # ------------------------------------
        # PromptManager (ipython >= 0.12)
        # ------------------------------------
        prompt = config.PromptManager
        prompt.in_template = '{DOOR_ALIAS} [\\#]: '
        prompt.in2_template = '   .\\D.: '
        prompt.out_template = 'Result [\\#]: '
        prompt.color_scheme = 'Linux'
    else:
        # (Deprecated in ipython >= 0.12 use PromptManager.in_template)
        i_shell.prompt_in1 = config.Spock.door_alias + ' [\\#]: '

        # (Deprecated in ipython >= 0.12 use PromptManager.in2_template)
        i_shell.prompt_in2 = '   .\\D.: '

        # (Deprecated in ipython >= 0.12 use PromptManager.out_template)
        i_shell.prompt_out = 'Result [\\#]: '

        # (Deprecated in ipython >= 0.12 use PromptManager.justify)
        i_shell.prompts_pad_left = True

    # ------------------------------------
    # IPCompleter
    # ------------------------------------
    completer = config.IPCompleter
    completer.omit__names = 2
    completer.greedy = False

    # ------------------------------------
    # InteractiveShellApp
    # ------------------------------------
    i_shell_app = config.InteractiveShellApp
    i_shell_app.ignore_old_config = True

    # ------------------------------------
    # TerminalIPythonApp: options for the IPython terminal (and not Qt Console)
    # ------------------------------------
    term_app = config.TerminalIPythonApp
    term_app.display_banner = True
    term_app.gui = gui_mode
    term_app.pylab='qt'
    #term_app.nosep = False
    #term_app.classic = True

    # ------------------------------------
    # IPKernelApp: options for the  Qt Console
    # ------------------------------------
    #kernel_app = config.IPKernelApp
    ipython_widget = config.IPythonWidget
    ipython_widget.in_prompt  = ' Spock [<span class="in-prompt-number">%i</span>]: '
    ipython_widget.out_prompt = 'Result [<span class="out-prompt-number">%i</span>]: '
    ipython_widget.input_sep = '\n'
    ipython_widget.output_sep = ''
    ipython_widget.output_sep2 = ''
    ipython_widget.enable_calltips = True
    if ipy_ver >= 1300:
        ipython_widget.gui_completion = 'droplist'
    else:
        ipython_widget.gui_completion = True
    ipython_widget.ansi_codes = True
    ipython_widget.paging = 'inside'
    #ipython_widget.pylab = 'inline'

    # ------------------------------------
    # ConsoleWidget
    # ------------------------------------
    # console_widget = config.ConsoleWidget

    # ------------------------------------
    # FrontendWidget
    # ------------------------------------
    frontend_widget = config.FrontendWidget
    frontend_widget.banner = banner

    # ------------------------------------
    # TerminalInteractiveShell
    # ------------------------------------
    term_i_shell = config.TerminalInteractiveShell
    term_i_shell.autocall = 2
    term_i_shell.automagic = True
    #term_i_shell.editor = 'gedit'
    #term_i_shell.editor = 'nano'

    term_i_shell.banner1 = banner
    term_i_shell.banner2 = "Connected to " + door_alias + "\n"
    #term_app.banner1 = banner
    #term_app.banner2 = "Connected to " + door_alias + "\n"

    # ------------------------------------
    # InlineBackend
    # ------------------------------------
    inline_backend = config.InlineBackend
    inline_backend.figure_format = 'svg'

    # ------------------------------------
    # InteractiveShellEmbed
    # ------------------------------------
    #i_shell_embed = config.InteractiveShellEmbed

    # ------------------------------------
    # NotebookApp
    # ------------------------------------
    #notebook_app = config.NotebookApp

    # ------------------------------------
    # NotebookManager
    # ------------------------------------
    #notebook_manager = config.NotebookManager

    # ------------------------------------
    # ZMQInteractiveShell
    # ------------------------------------
    #zmq_i_shell = config.ZMQInteractiveShell

    # Tell console everything is ready.
    config.Spock.ready = True
    return config

def start(user_ns=None):
    # Make sure the log level is changed to warning
    CodecFactory()
    taurus.setLogLevel(taurus.Warning)

    try:
        check_requirements()
    except exception.SpockMissingRequirement, requirement:
        print str(requirement)
        sys.exit(-1)
    except exception.SpockMissingRecommended, recommended:
        print str(recommended)

    user_ns = user_ns or {}
    try:
        user_ns.update(get_args(sys.argv))
    except exception.SpockException, e:
        print e.message
        print 'Starting normal IPython console'
    except KeyboardInterrupt:
        print "\nUser pressed Ctrl+C. Exiting..."
        sys.exit()
    except Exception, e:
        print 'spock exited with an unmanaged exception: %s' % str(e)
        sys.exit(-2)

    app = TerminalIPythonApp.instance()
    app.initialize()
    #config = get_config()
    return app

def mainloop(app=None, user_ns=None):
    if app is None:
        app = start(user_ns)
    app.start()

def prepare_input_handler():
    # initialize input handler as soon as possible
    import sardana.spock.inputhandler
    _ = sardana.spock.inputhandler.InputHandler()


def prepare_cmdline(argv=None):
    if argv is None:
        argv = sys.argv

    script_name = argv[0]
    _, session = os.path.split(script_name)
    script_name = os.path.realpath(script_name)

    # Define the profile file
    profile, append_profile = "spockdoor", True
    try:
        # in ipython the last option in the list takes precedence
        # so reversing order for searching of the profile 
        reversed_argv = reversed(argv[1:])
        for _, arg in enumerate(reversed_argv):
            if arg.startswith('--profile='):
                profile=arg[10:]
                append_profile = False
                break
    except:
        pass

    ipython_dir = get_ipython_dir()
    try:
        ProfileDir.find_profile_dir_by_name(ipython_dir, profile)
    except ProfileDirError:
        r = ''
        while not r in ('y', 'n'):
            prompt = "Profile '%s' does not exist. Do you want to create "\
                     "one now ([y]/n)? " % profile
            r = raw_input(prompt) or 'y'
        if r.lower() == 'y':
            create_spock_profile(ipython_dir, session, profile)
        else:
            sys.stdout.write('No spock profile was created. '
                             'Starting ipython with default profile...\n')
            sys.stdout.flush()
            # removing all options refering to profile
            for _, arg in enumerate(argv[1:]):
                if arg.startswith('--profile='):
                    argv.remove(arg)
            return

    if append_profile:
        argv.append("--profile=" + profile)


def run():

    try:
        from IPython.utils.traitlets import Unicode
        from IPython.qt.console.rich_ipython_widget import RichIPythonWidget
        
        class SpockConsole(RichIPythonWidget):

            banner = Unicode(config=True)

            def _banner_default(self):
                config = get_config()
                return config.FrontendWidget.banner

        import IPython.qt.console.qtconsoleapp
        IPythonQtConsoleApp = IPython.qt.console.qtconsoleapp.IPythonQtConsoleApp
        IPythonQtConsoleApp.widget_factory = SpockConsole
        IPythonQtConsoleApp.version.default_value = release.version
    except ImportError:
        pass

    try:
        check_requirements()
    except exception.SpockMissingRequirement, requirement:
        print str(requirement)
        sys.exit(-1)
    except exception.SpockMissingRecommended, recommended:
        print str(recommended)
    
    prepare_input_handler()
    prepare_cmdline()

    launch_new_instance()




#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import copy
import threading
import PyTango
from sardana.macroserver.macros.test import BaseMacroExecutor
from taurus.core.util.codecs import CodecFactory
from sardana import sardanacustomsettings


class TangoAttrCb(object):

    '''An abstract callback class for Tango events'''

    def __init__(self, tango_macro_executor):
        self._tango_macro_executor = tango_macro_executor


class TangoResultCb(TangoAttrCb):

    '''Callback class for Tango events of the Result attribute'''

    def push_event(self, *args, **kwargs):
        '''callback method receiving the event'''
        event_data = args[0]
        if event_data.err:
            result = event_data.errors
        else:
            result = event_data.attr_value.value
        self._tango_macro_executor._result = result


class TangoLogCb(TangoAttrCb):

    '''Callback class for Tango events of the log attributes
    e.g. Output, Error, Critical
    '''

    def __init__(self, tango_macro_executor, log_name):
        self._tango_macro_executor = tango_macro_executor
        self._log_name = log_name

    def push_event(self, *args, **kwargs):
        '''callback method receiving the event'''
        event_data = args[0]
        if event_data.attr_value:
            log = event_data.attr_value.value
            log_buffer_name = '_%s' % self._log_name
            log_buffer = getattr(self._tango_macro_executor, log_buffer_name)
            log_buffer.append(log)
            common_buffer = self._tango_macro_executor._common
            if common_buffer != None:
                common_buffer.append(log)


class TangoStatusCb(TangoAttrCb):

    '''Callback class for Tango events of the MacroStatus attribute'''

    START_STATES = ['start']
    DONE_STATES = ['finish', 'stop', 'exception']

    def push_event(self, *args, **kwargs):
        '''callback method receiving the event'''
        event_data = args[0]
        if event_data.err:
            self._state_buffer = event_data.errors
            self._tango_macro_executor._done_event.set()
        # make sure we get it as string since PyTango 7.1.4 returns a buffer
        # object and json.loads doesn't support buffer objects (only str)
        attr_value = event_data.attr_value
        v = map(str, attr_value.value)
        if not len(v[1]):
            return
        fmt = v[0]
        codec = CodecFactory().getCodec(fmt)

        # make sure we get it as string since PyTango 7.1.4 returns a buffer
        # object and json.loads doesn't support buffer objects (only str)
        v[1] = str(v[1])
        fmt, data = codec.decode(v)
        for macro_status in data:
            state = macro_status['state']
            self._tango_macro_executor._exception = macro_status.get(
                'exc_type')
            if state in self.START_STATES:
                #print 'TangoStatusCb.push_event: setting _started_event'
                self._tango_macro_executor._started_event.set()
            elif state in self.DONE_STATES:
                # print 'TangoStatusCb.push_event: setting _done_event %s'
                # %(state)
                self._tango_macro_executor._done_event.set()
            #else:
            #    print 'State %s' %(state)
            self._tango_macro_executor._state_buffer.append(state)


class TangoMacroExecutor(BaseMacroExecutor):

    '''
    Macro executor implemented using Tango communication with the Door device
    '''

    def __init__(self, door_name=None):
        super(TangoMacroExecutor, self).__init__()
        if door_name == None:
            door_name = getattr(sardanacustomsettings, 'UNITTEST_DOOR_NAME')
        self._door = PyTango.DeviceProxy(door_name)
        self._done_event = None
        self._started_event = None

    def _clean(self):
        '''Recreates threading Events in case the macro executor is reused.'''
        super(TangoMacroExecutor, self)._clean()
        self._started_event = threading.Event()
        self._done_event = threading.Event()

    def _run(self, macro_name, macro_params):
        '''reimplemented from :class:`BaseMacroExecutor`'''
        # preaparing RunMacro command input arguments
        argin = copy.copy(macro_params)
        argin.insert(0, macro_name)
        # registering for MacroStatus events
        status_cb = TangoStatusCb(self)
        self._status_id = self._door.subscribe_event('macrostatus',
                                                     PyTango.EventType.CHANGE_EVENT,
                                                     status_cb)
        # executing RunMacro command
        self._door.RunMacro(argin)

    def _wait(self, timeout):
        '''reimplemented from :class:`BaseMacroExecutor`'''
        #TODO: In case of timeout = inf if the macro excecutor run a macro
        # with wrong parameters it'll never awake of the done_event wait
        # Pending to remove this comment when Sardana resolves the bug.
        if self._done_event:
            self._done_event.wait(timeout)
            self._door.unsubscribe_event(self._status_id)

    def _stop(self, started_event_timeout=3.0):
        '''reimplemented from :class:`BaseMacroExecutor`'''
        self._started_event.wait(started_event_timeout)
        try:
            self._door.StopMacro()
        except PyTango.DevFailed as e:
            raise Exception("Unable to Stop macro: %s" % e)

    def _registerLog(self, log_level):
        '''reimplemented from :class:`BaseMacroExecutor`'''
        log_cb = TangoLogCb(self, log_level)
        id_log_name = '_%s_id' % log_level
        id_log = self._door.subscribe_event(log_level,
                                            PyTango.EventType.CHANGE_EVENT,
                                            log_cb)
        setattr(self, id_log_name, id_log)

    def _unregisterLog(self, log_level):
        '''reimplemented from :class:`BaseMacroExecutor`'''
        id_log_name = '_%s_id' % log_level
        id_log = getattr(self, id_log_name)
        self._door.unsubscribe_event(id_log)

    def _registerResult(self):
        '''reimplemented from :class:`BaseMacroExecutor`'''
        result_cb = TangoResultCb(self)
        self._result_id = self._door.subscribe_event('result',
                                                     PyTango.EventType.CHANGE_EVENT,
                                                     result_cb)

    def _unregisterResult(self):
        '''reimplemented from :class:`BaseMacroExecutor`'''
        self._door.unsubscribe_event(self._result_id)

    def getData(self):
        '''Returns the data object for the last executed macro

        :return: (obj)
        '''
        data = self._door.RecordData
        return CodecFactory().decode(data)



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""
delegate.py: 
"""

from taurus.external.qt import Qt

from sardana.taurus.core.tango.sardana import macro
from sardana.taurus.qt.qtgui.extra_macroexecutor.macroparameterseditor.parameditors import \
    AttrListComboBoxParam, SpinBoxParam, \
    DoubleSpinBoxParam, LineEditParam, FileDialogParam

from sardana.taurus.qt.qtgui.extra_macroexecutor import globals

class SequenceEditorDelegate(Qt.QItemDelegate):

    def __init__(self, parent=None):
        Qt.QItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        if index.column() == 2:
            macroNode = index.model().nodeFromIndex(index)
            opts = Qt.QStyleOptionProgressBarV2()
            opts.rect = option.rect
            range = macroNode.range()
            opts.minimum = range[0]
            opts.maximum = range[1]
            opts.textVisible = True
            percent = macroNode.progress()
            opts.progress = percent
#            opts.text = Qt.QString('Unavailable' if percent == 0 else '%d%%'%percent)
            opts.text = Qt.QString('%d%%' % percent)
#            opts.text = Qt.QString(percent)
            Qt.QApplication.style().drawControl(Qt.QStyle.CE_ProgressBar, opts, painter)
        else:
            Qt.QItemDelegate.paint(self, painter, option, index)


    def createEditor(self, parent, option, index):
        if index.column() == 3:
            return Qt.QCheckBox(parent)
        else:
            return Qt.QItemDelegate.createEditor(self, parent, option, index)

    def setEditorData(self, editor, index):
        node = index.model().mapToSource(index).internalPointer()
        if index.column() == 3:  #and isinstance(node, MacroNode):
            editor.setChecked(node.isPause())
        else:
            Qt.QItemDelegate.setEditorData(self, editor, index)

    def setModelData(self, editor, model, index):
        if index.column() == 3:
            model.setData(index, Qt.QVariant(editor.isChecked()))

class MacroParametersProxyDelegate(Qt.QItemDelegate):

    def __init__(self, parent=None):
        Qt.QItemDelegate.__init__(self, parent)

    def createEditor(self, parent, option, index):
        if index.column() == 1:
            node = index.model().nodeFromIndex(index)
            if isinstance(node, macro.SingleParamNode):
                paramType = node.type()
                if paramType in globals.EDITOR_COMBOBOX_PARAMS:
                    comboBox = AttrListComboBoxParam(parent, node)
                    comboBox.setUseParentModel(True)
                    return comboBox
                elif paramType in globals.EDITOR_SPINBOX_PARAMS:
                    return SpinBoxParam(parent, node)
                elif paramType in globals.EDITOR_DOUBLESPINBOX_PARAMS:
                    return DoubleSpinBoxParam(parent, node)
                elif paramType in globals.EDITOR_LINEEDIT_PARAMS:
                    return LineEditParam(parent, node)
                elif paramType in globals.EDITOR_FILEDIALOG_PARAMS:
                    return FileDialogParam(parent, node)
        return Qt.QItemDelegate.createEditor(self, parent, option, index)

    def setEditorData(self, editor, index):
        if index.column() == 1:
            text = Qt.from_qvariant(index.model().data(index, Qt.Qt.DisplayRole), str)
            if text == "None" or text == "":
                pass
            else:
                node = index.model().mapToSource(index).internalPointer()
                paramType = node.type()
                if paramType in globals.EDITOR_COMBOBOX_PARAMS :
                    i = editor.findText(text)
                    if i == -1:
                        i = 0
                    else:
                        editor.previous = text
                    editor.setCurrentIndex(i)
                elif paramType in globals.EDITOR_SPINBOX_PARAMS:
                    editor.setValue(int(text))
                elif paramType in globals.EDITOR_DOUBLESPINBOX_PARAMS:
                    editor.setValue(float(text))
                elif paramType in globals.EDITOR_LINEEDIT_PARAMS:
                    editor.setText(text)
                elif paramType in globals.EDITOR_FILEDIALOG_PARAMS:
                    editor.filePath.setText(text)
        else:
            Qt.QItemDelegate.setEditorData(self, editor, index)

    def setModelData(self, editor, model, index):
        if index.column() == 1:
            model.setData(index, Qt.QVariant(editor.getValue()))

    def sizeHint(self, option, index):
        if index.column() == 0:
            fm = option.fontMetrics
            text = Qt.from_qvariant(index.model().data(index, Qt.Qt.DisplayRole), str)
            document = Qt.QTextDocument()
            document.setDefaultFont(option.font)
            document.setHtml(text)
            size = Qt.QSize(document.idealWidth() + 5, fm.height())
        elif index.column() == 1:
            editor = self.createEditor(self.parent(), option, index)
            if editor is None:
                size = Qt.QItemDelegate.sizeHint(self, option, index)
            else:
                size = editor.sizeHint()
                editor.hide()
                editor.setParent(None)
#                editor.destroy()
        else:
            size = Qt.QItemDelegate.sizeHint(self, option, index)
        return size





#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This is the macro server scan data output recorder module"""

__all__ = ["JsonRecorder", "OutputRecorder"]

__docformat__ = 'restructuredtext'

import numpy
import datetime
import operator
import string

from taurus.core.util.codecs import CodecFactory
from taurus.core.util.containers import CaselessList

from sardana.macroserver.scan.recorder.datarecorder import DataRecorder
from sardana.macroserver.scan.recorder.storage import BaseFileRecorder


class JsonRecorder(DataRecorder):
    def __init__(self, stream, cols=None, **pars):
        DataRecorder.__init__(self, **pars)
        self._stream = stream
        self._codec = CodecFactory().getCodec('json')

    def _startRecordList(self, recordlist):
        macro_id = recordlist.getEnvironValue('macro_id')
        title = recordlist.getEnvironValue('title')
        counters = recordlist.getEnvironValue('counters')
        scanfile = recordlist.getEnvironValue('ScanFile')
        scandir = recordlist.getEnvironValue('ScanDir')
        serialno = recordlist.getEnvironValue('serialno')
        column_desc = recordlist.getEnvironValue('datadesc')
        ref_moveables = recordlist.getEnvironValue('ref_moveables')
        estimatedtime = recordlist.getEnvironValue('estimatedtime')
        total_scan_intervals = recordlist.getEnvironValue('total_scan_intervals')
        start_time = recordlist.getEnvironValue('starttime').ctime()
        self.column_desc = []
        discarded = []
        for e in column_desc:
            if len(e.shape) == 0:
                self.column_desc.append(e)
            else:
                discarded.append(e.label)
        if discarded:
            self.info('The following data will not be json-serialized: %s',
                      " ".join(discarded))
        column_desc = [d.toDict() for d in self.column_desc]
        data = {'column_desc': column_desc,
                'ref_moveables': ref_moveables,
                'estimatedtime': estimatedtime,
                'total_scan_intervals': total_scan_intervals,
                'starttime': start_time,
                'title': title,
                'counters': counters,
                'scanfile': scanfile,
                'scandir': scandir,
                'serialno': serialno}
        self._sendPacket(type="data_desc", data=data, macro_id=macro_id)

    def _endRecordList(self, recordlist):
        macro_id = recordlist.getEnvironValue('macro_id')
        data = { 'endtime'  : recordlist.getEnvironValue('endtime').ctime(),
                 'deadtime' : recordlist.getEnvironValue('deadtime') }
        self._sendPacket(type="record_end", data=data, macro_id=macro_id)

    def _writeRecord(self, record):
        macro_id = self.recordlist.getEnvironValue('macro_id')
        data = {}  # dict(record.data)
        for k in self.column_desc:
            name = k.name
            data[name] = record.data[name]
        self._sendPacket(type="record_data", data=data, macro_id=macro_id)

    def _sendPacket(self, **kwargs):
        '''creates a JSON packet using the keyword arguments passed
        and then sends it'''
        #data = self._codec.encode(('', kwargs))
        #self._stream.sendRecordData(*data)
        self._stream.sendRecordData(kwargs, codec='json')

    def _addCustomData(self, value, name, **kwargs):
        '''
        The custom data will be sent as a packet with type='custom_data'
        and its data will be the dictionary of keyword arguments passed to this
        method plus 'name' and 'value'
        '''
        #try to convert to list to avoid serialization problems
        try:
            value = value.tolist()
        except:
            pass
        macro_id = self._stream.getID()
        data = dict(kwargs)  # shallow copy
        data['name'] = name
        data['value'] = value
        self._sendPacket(type="custom_data", data=data, macro_id=macro_id)


class OutputRecorder(DataRecorder):

    def __init__(self, stream, cols=None, number_fmt='%8.4f', col_width=8,
                 col_sep='  ', **pars):
        DataRecorder.__init__(self, **pars)
        self._stream = stream
        if not number_fmt.startswith('%'):
            number_fmt = '%%s' % number_fmt
        self._number_fmt = number_fmt
        self._col_sep = col_sep
        self._col_width = col_width
        if operator.isSequenceType(cols) and \
                not isinstance(cols, (str, unicode)):
            cols = CaselessList(cols)
        elif operator.isNumberType(cols):
            cols = cols
        else:
            cols = None
        self._columns = cols

    def _startRecordList(self, recordlist):
        starttime = recordlist.getEnvironValue('starttime').ctime()
        estimatedtime = recordlist.getEnvironValue('estimatedtime')
        data_desc = recordlist.getEnvironValue('datadesc')
        serialno = recordlist.getEnvironValue('serialno')
        col_sep = self._col_sep
        cols = self._columns
        number_fmt = self._number_fmt
        col_width = self._col_width
        dh = recordlist.getDataHandler()

        for fr in [r for r in dh.recorders if isinstance(r, BaseFileRecorder)]:
            self._stream.info('Operation will be saved in %s (%s)',
                              fr.getFileName(), fr.getFormat())

        msg = "Scan #%d started at %s." % (serialno, starttime)
        if not estimatedtime is None:
            estimatedtime = datetime.timedelta(0, abs(estimatedtime))
            msg += " It will take at least %s\n" % estimatedtime
            msg += "Moving to start positions..."
        self._stream.info(msg)

        labels, col_names, col_sizes = [], [], []
        header_rows, header_len = 1, 0
        for col, column in enumerate(data_desc):
            if not getattr(column, 'output', True):
                continue
            name = column.name
            if operator.isSequenceType(cols) and name not in cols:
                continue
            if operator.isNumberType(cols) and col >= cols:
                break
            col_names.append(name)
            label = column.label.strip()
            if len(label) > col_width:
                label = label.split("/")
            else:
                label = [label]
            header_rows = max(header_rows, len(label))
            labels.append(label)
            col_size = max(col_width, max(map(len, label)))
            header_len += col_size
            col_sizes.append(col_size)

        nb_cols = len(col_names)
        header_len += (nb_cols - 1) * len(col_sep)
        self._labels = labels
        self._col_names = col_names
        self._col_sizes = col_sizes

        header = [[] for i in range(header_rows)]
        for col, (label, col_size) in enumerate(zip(labels, col_sizes)):
            empty_row_nb = header_rows - len(label)
            for row in range(empty_row_nb):
                header[row].append(col_size * " ")
            for i, l in enumerate(label):
                header[i + empty_row_nb].append(string.center(l, col_size))
        head = []
        for header_row in header:
            head.append(col_sep.join(header_row))

        header = "\n".join(head)

        cell_t_number = '%%%%(%%s)%s' % number_fmt[1:]

        self._scan_line_t = [(col_names[0], '%%(%s)8d' % col_names[0])]
        self._scan_line_t += [(name, cell_t_number % name) for name in col_names[1:]]

        self._stream.output(header)
        self._stream.flushOutput()

    def _endRecordList(self, recordlist):
        self._stream.flushOutput()
        starttime = recordlist.getEnvironValue('starttime')
        endtime = recordlist.getEnvironValue('endtime')
        deadtime = recordlist.getEnvironValue('deadtime')
        motiontime = recordlist.getEnvironValue('motiontime')
        totaltime = endtime - starttime
        endtime = endtime.ctime()
        serialno = recordlist.getEnvironValue('serialno')

        dh = recordlist.getDataHandler()

        for fr in [ r for r in dh.recorders if isinstance(r, BaseFileRecorder) ]:
            self._stream.info('Operation saved in %s (%s)', fr.getFileName(),
                              fr.getFormat())

        endts = recordlist.getEnvironValue('endts')
        startts = recordlist.getEnvironValue('startts')
        totaltimets = endts - startts
        deadtime_perc = deadtime * 100.0 / totaltimets
        motiontime_perc = motiontime * 100.0 / totaltimets
        info_string = 'Scan #%s ended at %s, taking %s.' + \
                      'Dead time %.1f%% (motion dead time %.1f%%)'
        self._stream.info(info_string % (serialno, endtime, totaltime,
                                         deadtime_perc, motiontime_perc))

    def _writeRecord(self, record):
        cells = []
        for i, (name, cell) in enumerate(self._scan_line_t):
            cell_data = record.data[name]
            if isinstance(cell_data, numpy.ndarray):
                cell = str(cell_data.shape)
            elif cell_data is None:
                cell = "<nodata>"
            elif isinstance(cell_data, (str, unicode)):
                cell = "<string>"
            else:
                cell %= record.data
            cell = string.center(cell.strip(), self._col_sizes[i])
            cells.append(cell)
        scan_line = self._col_sep.join(cells)

        self._stream.output(scan_line)
        self._stream.flushOutput()

    def _addCustomData(self, value, name, **kwargs):
        '''
        The custom data will be added as an info line in the form:
        Custom data: name : value
        '''
        if numpy.rank(value) > 0:
            v = 'Array(%s)' % str(numpy.shape(value))
        else:
            v = str(value)
        self._stream.output('Custom data: %s : %s' % (name, v))
        self._stream.flushOutput()



##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains macros that demonstrate the usage of macro parameters"""

from sardana.macroserver.macro import *

__all__ = ["pt0", "pt1", "pt2", "pt3", "pt4", "pt5", "pt6", "pt7", "pt8",
           "pt9"]

class pt0(Macro):
    """Macro without parameters. Pretty dull"""

    param_def = []
    
    def run(self):
        pass
    
class pt1(Macro):
    """Macro with one float parameter: Each parameter is described in the 
    param_def sequence as being a sequence of for elements: name, type, 
    default value and description"""
    
    param_def = [ [ 'value', Type.Float, None, 'some bloody float'] ]
    
    def run(self, f):
        pass

class pt2(Macro):
    """Macro with one Motor parameter: Each parameter is described in the 
    param_def sequence as being a sequence of for elements: name, type, 
    default value and description"""

    param_def = [ [ 'motor', Type.Motor, None, 'some bloody motor'] ]
    
    def run(self, m):
        pass

class pt3(Macro):
    """Macro with a list of numbers as parameter: the type is a sequence of
    parameter types which is repeated. In this case it is a repetition of a 
    float so only one parameter is defined.
    By default the repetition as a semantics of 'at least one'"""

    param_def = [
       [ 'numb_list', [ [ 'pos', Type.Float, None, 'value'] ], None, 'List of values'],
    ]
    
    def run(self, *args, **kwargs):
        pass

class pt4(Macro):
    """Macro with a list of motors as parameter: the type is a sequence of
    parameter types which is repeated. In this case it is a repetition of a 
    motor so only one parameter is defined.
    By default the repetition as a semantics of 'at least one'"""

    param_def = [
       [ 'motor_list', [ [ 'motor', Type.Motor, None, 'motor name'] ], None, 'List of motors'],
    ]
    
    def run(self, *args, **kwargs):
        pass

class pt5(Macro):
    """Macro with a motor parameter followed by a list of numbers"""

    param_def = [
       [ 'motor', Type.Motor, None, 'Motor to move'],
       [ 'numb_list', [ [ 'pos', Type.Float, None, 'value'] ], None, 'List of values'],
    ]
    
    def run(self, *args, **kwargs):
        pass

class pt6(Macro):
    """Macro with a motor parameter followed by a list of numbers. The list as
    explicitly stated an optional last element which is a dictionary that defines the
    min and max values for repetitions"""

    param_def = [
       [ 'motor', Type.Motor, None, 'Motor to move'],
       [ 'numb_list', [ [ 'pos', Type.Float, None, 'value'], { 'min' : 1, 'max' : None } ], None, 'List of values'],
    ]
    
    def run(self, *args, **kwargs):
        pass
    
class pt7(Macro):
    """Macro with a list of pair Motor,Float"""

    param_def = [
       [ 'm_p_pair', [ [ 'motor', Type.Motor, None, 'Motor to move'],
                       [ 'pos',   Type.Float, None, 'Position to move to'] ],
         None, 'List of motor/position pairs']
    ]
    
    def run(self, *args, **kwargs):
        pass

class pt8(Macro):
    """Macro with a list of pair Motor,Float. The min and max elements have been
    explicitly stated"""
    
    param_def = [
       [ 'm_p_pair', [ [ 'motor', Type.Motor, None, 'Motor to move'],
                       [ 'pos',   Type.Float, None, 'Position to move to'],
                       { 'min' : 1, 'max' : 2 } ],
         None, 'List of motor/position pairs']
    ]
    
    def run(self, *args, **kwargs):
        pass

class pt9(Macro):
    """Same as macro pt7 but witb old style ParamRepeat. If you are writing
    a macro with variable number of parameters for the first time don't even
    bother to look at this example since it is DEPRECATED."""

    param_def = [
       ['m_p_pair',
        ParamRepeat(['motor', Type.Motor, None, 'Motor to move'],
                    ['pos',  Type.Float, None, 'Position to move to'], min=1, max= 2),
        None, 'List of motor/position pairs'],
    ]
    
    def run(self, *args, **kwargs):
        pass
    
class twice(Macro):
    """A macro that returns a float that is twice its input. It also sets its 
    data to be a dictionary with 'in','out' as keys and value,result 
    as values, respectively"""

    # uncomment the following lines as necessary. Otherwise you may delete them
    param_def = [ [ "value", Type.Float, 23, "value to be doubled" ] ]
    result_def = [ [ "result", Type.Float, 23, "the double of the given value" ] ]
    #hints = {}
    #env = (,)
    
    # uncomment the following lines if need prepare. Otherwise you may delete them
    #def prepare(self):
    #    pass
        
    def run(self, n):
        ret = 2*n
        self.setData({'in':n, 'out':ret})
        return ret


#!/usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""Initial magic commands and hooks for the spock IPython environment"""

__all__ = ['expconf', 'showscan', 'spsplot', 'debug_completer',
           'debug', 'www',
           'post_mortem', 'macrodata', 'edmac', 'spock_late_startup_hook',
           'spock_pre_prompt_hook']


from .genutils import page, get_door, get_macro_server, ask_yes_no, arg_split
from .genutils import MSG_DONE, MSG_FAILED
from .genutils import get_ipapi


def expconf(self, parameter_s=''):
    """Launches a GUI for configuring the environment variables
    for the experiments (scans)"""
    try:
        from sardana.taurus.qt.qtgui.extra_sardana import ExpDescriptionEditor
    except:
        print "Error importing ExpDescriptionEditor "\
              "(hint: is taurus extra_sardana installed?)"
        return
    doorname = get_door().name()
        
    #===========================================================================
    ## ugly hack to avoid ipython/qt thread problems #e.g. see
    ## https://sourceforge.net/p/sardana/tickets/10/ 
    ## this hack does not allow inter-process communication and leaves the 
    ## widget open after closing spock 
    ## @todo: investigate cause of segfaults when using launching qt widgets from ipython
    # 
    #w = ExpDescriptionEditor(door=doorname)
    #w.show() #launching it like this, produces the problem of https://sourceforge.net/p/sardana/tickets/10/
    import subprocess
    import sys
    fname = sys.modules[ExpDescriptionEditor.__module__].__file__
    args = ['python', fname, doorname]
    subprocess.Popen(args)
    #===========================================================================


def showscan(self, parameter_s=''):
    """Shows a scan in a GUI.

    :param scan_id: scan number [default: None, meaning show last scan]"""
    params = parameter_s.split()
    door = get_door()
    online, scan_nb = False, None
    if len(params) > 0:
        if params[0].lower() == 'online':
            msg = 'To see the scans online, launch "expconf" and ' + \
                  'enable the plots from the "plots" button ' + \
                  '(top-right in the first tab)'
            print msg
            return
        # show the scan plot, ignoring the plot configuration
        elif params[0].lower() == 'online_raw':
            online = True
        else:
            scan_nb = int(params[0])
    door.show_scan(scan_nb, online=online)


def spsplot(self, parameter_s=''):
    get_door().plot()


def debug_completer(self, event):
    # calculate parameter index
    param_idx = len(event.line.split()) - 1
    if not event.line.endswith(' '):
        param_idx -= 1
    if param_idx == 0:
        return ('off', 'on')


def debug(self, parameter_s=''):
    """Activate/Deactivate macro server debug output"""
    params = parameter_s.split()
    door = get_door()
    if len(params) == 0:
        s = door.getDebugMode() and 'on' or 'off'
        print "debug mode is %s" % s
        return
    elif len(params) == 1:
        s = params[0].lower()
        if not s in ('off', 'on'):
            print "Usage: debug [on|off]"
            return
        door.setDebugMode(s == 'on')
        print "debug mode is now %s" % s
    else:
        print "Usage: debug [on|off]"


def www(self, parameter_s=''):
    """What went wrong. Prints the error message from the last macro execution"""
    import PyTango
    door = get_door()
    try:
        last_macro = door.getLastRunningMacro()
        if last_macro is None:
            door.writeln("No macro ran from this console yet!")
            return
        if not hasattr(last_macro, 'exc_stack') or last_macro.exc_stack is None:
            door.writeln("Sorry, but no exception occurred running last " \
                         "macro (%s)." % last_macro.name)
            return
        exc = "".join(last_macro.exc_stack)
        door.write(exc)
    except Exception, e:
        door.writeln("Unexpected exception occurred executing www:",
                     stream=door.Error)
        door.writeln(str(e), stream=door.Error)
        import traceback
        traceback.print_exc()


def post_mortem(self, parameter_s='', from_www=False):
    """Post mortem analysis. Prints the local stream buffer. If no stream is
    specified, it reads 'debug' stream. Valid values are output, critical,
    error, warning, info, debug, result"""
    params = parameter_s.split() or ['debug']
    door = get_door()
    logger = door.getLogObj(params[0])
    msg = ""

    if not from_www:
        try:
            msg = "\n".join(logger.read(cache=False).value)
        except:
            from_www = True
            
    if from_www:
        msg = "------------------------------\n" \
              "Server is offline.\n" \
              "This is a post mortem analysis\n" \
              "------------------------------\n"
        msg += "\n".join(logger.getLogBuffer())
    page(msg)

def macrodata(self, parameter_s=''):
    """macrodata
    
    Returns the data produced by the last macro"""
    door = get_door()
    macro_data = door.read_attribute("RecordData")

    from taurus.core.util.codecs import CodecFactory
    factory = CodecFactory()
    data = factory.decode(macro_data.value)    
    return data
    

def edmac(self, parameter_s=''):
    """edmac <macro name> [<module>]
    Returns the contents of the macro file which contains the macro code for
    the given macro name. If the module is given and it does not exist a new
    one is created. If the given module is a simple module name and it does
    not exist, it will be created on the first directory mentioned in the
    MacroPath"""
    import os
    import tempfile
    import PyTango

    ms = get_macro_server()

    pars = arg_split(parameter_s, posix=True)

    if len(pars) == 1:
        macro_name = pars[0]
        macro_info = ms.getMacroInfoObj(macro_name)
        if macro_info is None:
            print "Macro '%s' could not be found" % macro_name
            return
        macro_lib = macro_info.module
    else:
        macro_name, macro_lib = pars

    macro_info = (macro_lib, macro_name)
    print 'Opening %s.%s...' % macro_info

    try:
        remote_fname, code, line_nb = ms.GetMacroCode(macro_info)
    except PyTango.DevFailed, e:
        PyTango.Except.print_exception(e)
        return

    fd, local_fname = tempfile.mkstemp(prefix='spock_%s_' % pars[0],
                                       suffix='.py', text=True)
    os.write(fd, code)
    os.close(fd)

    cmd = 'edit -x -n %s %s' % (line_nb, local_fname)
    ip = get_ipapi()
    ip.magic(cmd)

    if ask_yes_no('Do you want to apply the new code on the server?', 'y'):
        print 'Storing...',
        try:
            f = file(local_fname)
            try:
                new_code = f.read()
                ms.SetMacroCode([remote_fname, new_code])
                print MSG_DONE
            except Exception, e:
                print MSG_FAILED
                print 'Reason:', str(e)
            f.close()
        except:
            print 'Could not open file \'%s\' for safe transfer to the ' \
                  'server' % local_fname
            print 'Did you forget to save?'
    else:
        print "Discarding changes..."

    #if os.path.exists(local_fname):
    #    if ask_yes_no('Delete temporary file \'%s\'?' % local_fname, 'y'):
    #        os.remove(local_fname)
    #        bkp = '%s~' % local_fname
    #        if os.path.exists(bkp):
    #            os.remove(bkp)
    try:
        os.remove(local_fname)
    except:
        pass


def spock_late_startup_hook(self):
    try:
        get_door().setConsoleReady(True)
    except:
        import traceback
        print "Exception in spock_late_startup_hook:"
        traceback.print_exc()


def spock_pre_prompt_hook(self):
    try:
        get_door().pre_prompt_hook(self)
    except:
        import traceback
        print "Exception in spock_pre_prompt_hook:"
        traceback.print_exc()

#def spock_pre_runcode_hook(self):
#    print "spock_pre_runcode_hook"
#    return None



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import PyTango

import taurus
from taurus.external.qt import Qt
from taurus.core.taurusbasetypes import TaurusEventType
from taurus.qt.qtgui.base import TaurusBaseWidget
from taurus.qt.qtgui.input import TaurusAttrListComboBox
from taurus.qt.qtgui.container import TaurusMainWindow
from taurus.qt.qtgui.resource import getThemeIcon


def str2bool(text):
    return text in ("True", "1")

def standardPlotablesFilter(column_desc_dict):
    """This is a standard plotables filter emitted with "plotablesFilter" signal.
       Its purpose is to filterout point_nr columnf from scanline"""
    if column_desc_dict.get("label", None) == "point_nb":
        return False
    return True


class MSAttrListComboBox(TaurusAttrListComboBox):
    _elementType = ''

    def handleEvent(self, evt_src, evt_type, evt_value):
        if evt_type in (TaurusEventType.Config, TaurusEventType.Error):
            return
        text = str(self.currentText())
        self.clear()
        parentModelObj = self.getParentModelObj()
        if parentModelObj == None:
            return
        items = parentModelObj.getElementNamesWithInterface(self._elementType)
        items.sort()
        self.addItems(items)
        if text in items:
            self.setCurrentText(text)

    def setCurrentText(self, text):
        idx = self.findText(text)
        self.setCurrentIndex(idx)

    def setElementType(self, elementType):
        self._elementType = elementType

    def getElementType(self):
        return self._elementType

    def resetElementType(self):
        self._elementType = MSAttrListComboBox._elementType

    elementType = Qt.pyqtProperty("QString", getElementType, setElementType, resetElementType)


class MacroComboBox(Qt.QComboBox, TaurusBaseWidget):
    """ComboBox representing list of macros"""


    def __init__(self, parent=None):
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QComboBox, parent)
        self.call__init__(TaurusBaseWidget, name)
        self.setSizeAdjustPolicy(Qt.QComboBox.AdjustToContentsOnFirstShow)
        self.setToolTip("Choose a macro name...")

    def setModel(self, model):
        if isinstance(model, Qt.QAbstractItemModel):
            Qt.QAbstractItemView.setModel(self, model)
        else:
            TaurusBaseWidget.setModel(self, model)
            self.__loadMacroNames()

    def parentModelChanged(self, parentmodel_name):
        TaurusBaseWidget.parentModelChanged(self, parentmodel_name)
        self.__loadMacroNames()

    def onMacrosUpdated(self):
        self.__loadMacroNames()

    def __loadMacroNames(self):
        self.clear()
        ms = self.getModelObj()
        if ms == None: return
        macros = ms.getElementsWithInterface('MacroCode')
        macroNames = [macro.name for macro in macros.values()]
        macroNames.sort()
        macroNames.insert(0, '')  #adding blank item
        self.addItems(macroNames)
        self.updateStyle()

    def selectMacro(self, macroName):
        currentIdx = self.currentIndex()
        index = self.findText(macroName)
        self.setCurrentIndex(index)
        if currentIdx == index:
            self.emit(Qt.SIGNAL("currentIndexChanged(QString)"), macroName)


class TaurusMacroConfigurationDialog(Qt.QDialog):

    def __init__(self, parent=None, initMacroServer=None, initDoor=None):
        Qt.QDialog.__init__(self, parent)
        self.initMacroServer = initMacroServer
        self.initDoor = initDoor
        configureAction = Qt.QAction(getThemeIcon("folder-open"), "Change custom macro editors paths", self)
        self.connect(configureAction, Qt.SIGNAL("triggered()"), self.onReloadMacroServers)
        configureAction.setToolTip("Change custom macro editors paths")
        configureAction.setShortcut("F11")
        self.refreshMacroServersAction = Qt.QAction(getThemeIcon("view-refresh"), "Reload macroservers", self)
        self.connect(self.refreshMacroServersAction, Qt.SIGNAL("triggered()"), self.onReloadMacroServers)
        self.refreshMacroServersAction.setToolTip("This will reload list of all macroservers from Tango DB")
        self.refreshMacroServersAction.setShortcut("F5")
        self.initComponents()

    def initComponents(self):
        self.setModal(True)
        macroServerLabel = Qt.QLabel("MacroServer:", self)
        self.macroServerComboBox = Qt.QComboBox()
        ms_stateIcons = self.__retriveMacroServersFromDB()
        self.__fillMacroServerComboBox(ms_stateIcons, self.macroServerComboBox)
        refreshMacroServersButton = Qt.QToolButton()
        refreshMacroServersButton.setDefaultAction(self.refreshMacroServersAction)
        doorLabel = Qt.QLabel("Door:", self)
        self.doorComboBox = TaurusAttrListComboBox(self)
        self.doorComboBox.setModel(self.macroServerComboBox.currentText() + "/doorList")

        self.buttonBox = Qt.QDialogButtonBox(Qt.QDialogButtonBox.Ok |
                                                Qt.QDialogButtonBox.Cancel)
        gridLayout = Qt.QGridLayout()
        gridLayout.addWidget(macroServerLabel, 0, 0)
        gridLayout.addWidget(self.macroServerComboBox, 0, 1)
        gridLayout.addWidget(refreshMacroServersButton, 0, 2)
        gridLayout.addWidget(doorLabel, 1, 0)
        gridLayout.addWidget(self.doorComboBox, 1, 1)

        self.setLayout(Qt.QVBoxLayout())
        self.layout().addLayout(gridLayout)
        self.layout().addWidget(self.buttonBox)
        self.adjustSize()

        self.connect(self.buttonBox, Qt.SIGNAL("accepted()"), self, Qt.SLOT("accept()"))
        self.connect(self.buttonBox, Qt.SIGNAL("rejected()"), self, Qt.SLOT("reject()"))
        self.connect(self.macroServerComboBox, Qt.SIGNAL("currentIndexChanged(const QString&)"), self.onMacroServerComboBoxChanged)
        self.selectMacroServer(self.initMacroServer)
        self.selectDoor(self.initDoor)

    def accept(self):
        self.emit(Qt.SIGNAL("macroserverNameChanged"), str(self.macroServerComboBox.currentText()))
        self.emit(Qt.SIGNAL("doorNameChanged"), str(self.doorComboBox.currentText()))
        Qt.QDialog.accept(self)

    def __retriveMacroServersFromDB(self):
        ms_stateIcons = []
        db = taurus.Database()
        macroServerList = db.getValueObj().get_device_name('*', 'MacroServer')
        for macroServer in macroServerList:
            #state = Device(macroServer).getState()
            state = None
            try:
                state = PyTango.DeviceProxy(macroServer).state()
            except:
                pass
            icon = None
            if state == PyTango.DevState.ON:
                icon = Qt.QIcon(":/leds/images24/ledgreen.png")
            elif state == PyTango.DevState.FAULT:
                icon = Qt.QIcon(":/leds/images24/ledred.png")
            elif state == None:
                icon = Qt.QIcon(":/leds/images24/ledredoff.png")
            ms_stateIcons.append((macroServer, icon))
        return ms_stateIcons

    def onReloadMacroServers(self):
        ms_stateIcons = self.__retriveMacroServersFromDB()
        self.__fillMacroServerComboBox(ms_stateIcons, self.macroServerComboBox)

    def onMacroServerComboBoxChanged(self, macroServerName):
        self.doorComboBox.setModel(macroServerName + "/doorList")
        self.doorComboBox.fireEvent(self.doorComboBox, taurus.core.taurusbasetypes.TaurusEventType.Change, self.doorComboBox.getModelValueObj())  #fake event

    def onMacroServerNameChanged(self, macroServerName):
        self.__selectMacroServer(macroServerName)

    def onDoorNameChanged(self, doorName):
        self.__selectDoor(doorName)

    def __fillMacroServerComboBox(self, ms_stateIcons, comboBox):
        comboBox.clear()
        for ms_stateIcon in ms_stateIcons:
            macroServer = ms_stateIcon[0]
            icon = ms_stateIcon[1]
            comboBox.addItem(icon, macroServer)

    def selectDoor(self, doorName):
        if doorName is None:
            return
        #@todo: Change that it will be able to handle also full device names
        "/".join(doorName.split("/")[-3:])
        index = self.doorComboBox.findText(doorName)
        if index != -1:
            self.doorComboBox.setCurrentIndex(index)

    def selectMacroServer(self, macroServerName):
        if macroServerName is None:
            return
        #@todo: Change that it will be able to handle also full device names
        "/".join(macroServerName.split("/")[-3:])
        index = self.macroServerComboBox.findText(macroServerName)
        if index != -1:
            self.macroServerComboBox.setCurrentIndex(index)

class MacroExecutionWindow(TaurusMainWindow):

    def __init__(self, parent=None, designMode=False):
        TaurusMainWindow.__init__(self, parent, designMode)
        self.statusBar().showMessage("")
        self.setModelInConfig(True)
        self._doorName = ""
        self.registerConfigProperty("doorName", "setDoorName", "doorName")
        self._customMacroEditorPaths = ""
        self.registerConfigProperty("customMacroEditorPaths", "setCustomMacroEditorPaths", "customMacroEditorPaths")
        self._qDoor = None
        self.setWindowIcon(Qt.QIcon(":/apps/preferences-system-session.svg"))
        toolBar = self.basicTaurusToolbar()
        toolBar.setIconSize(Qt.QSize(24, 24))
        self.configureAction = self.createConfigureAction()
        toolBar.addAction(self.configureAction)
        self.taurusMenu.addAction(self.configureAction)
        self.customMacroEditorsPathsAction = self.createCustomMacroEditorPathsAction()
        self.taurusMenu.addAction(self.customMacroEditorsPathsAction)
        self.addToolBar(toolBar)
        self.initComponents()
        self.splashScreen().finish(self)
        self.connect(self, Qt.SIGNAL("doorChanged"), self.onDoorChanged)

    def doorName(self):
        return self._doorName

    def setDoorName(self, doorName):
        self._doorName = doorName

    def onDoorChanged(self, doorName):
        self.setDoorName(doorName)

    def customMacroEditorPaths(self):
        return self._customMacroEditorPaths

    def setCustomMacroEditorPaths(self, customMacroEditorPaths):
        self._customMacroEditorPaths = customMacroEditorPaths
#        ParamEditorManager().parsePaths(customMacroEditorPaths)
#        ParamEditorManager().browsePaths()

    def onCustomMacroEditorPaths(self):
        paths = str(Qt.QInputDialog.getText(self,
                                "Edition of custom macro editors paths",
                                "Paths:", Qt.QLineEdit.Normal,
                                str(self.customMacroEditorPaths()))[0])
        self.setCustomMacroEditorPaths(paths)

    def initComponents(self):
        pass

    def setModel(self, model):
        """Sets new model for application, and change window title witn new macroserver name."""
        TaurusMainWindow.setModel(self, model)
        self.setWindowTitle(Qt.QApplication.applicationName() + ": " + model)

    def createConfigureAction(self):
        configureAction = Qt.QAction(getThemeIcon("preferences-system-session"), "Change configuration", self)
        self.connect(configureAction, Qt.SIGNAL("triggered()"), self.changeConfiguration)
        configureAction.setToolTip("Configuring MacroServer and Door")
        configureAction.setShortcut("F10")
        return configureAction

    def createCustomMacroEditorPathsAction(self):
        configureAction = Qt.QAction(getThemeIcon("folder-open"), "Change custom macro editors paths", self)
        self.connect(configureAction, Qt.SIGNAL("triggered()"), self.onCustomMacroEditorPaths)
        configureAction.setToolTip("Change custom macro editors paths")
        configureAction.setShortcut("F11")
        return configureAction

    def changeConfiguration(self):
        """This method is used to change macroserver as a model of application.
           It shows dialog with list of all macroservers on tango host, if the user
           Cancel dialog it doesn't do anything."""
        dialog = TaurusMacroConfigurationDialog(self, self.modelName, self.doorName())
        if dialog.exec_():
            self.setModel(str(dialog.macroServerComboBox.currentText()))
            self.emit(Qt.SIGNAL("doorChanged"), str(dialog.doorComboBox.currentText()))
        else:
            return

    def onShortMessage(self, msg):
        ''' Slot to be called when there is a new short message. Currently, the only action 
        taken when there is a new message is to display it in the main window status bar.
        
        :param msg: (str) the short descriptive message to be handled 
        '''
        self.statusBar().showMessage(msg)

def test_macrocombobox(ms_name):
    mcb = MacroComboBox()
    mcb.setModel(ms_name)
    mcb.show()

if __name__ == "__main__":
    import sys
    from taurus.qt.qtgui.application import TaurusApplication
    app = TaurusApplication()
    args = app.get_command_line_args()
    ms_name = args[0]
    test_macrocombobox(ms_name)
    sys.exit(app.exec_())




#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""System tests for Macros"""

__all__ = ['macroTest', 'BaseMacroTestCase', 'RunMacroTestCase',
           'RunStopMacroTestCase', 'testRun', 'testFail', 'testStop']
import time
import functools
from sardana import sardanacustomsettings
from sardana.macroserver.macros.test import MacroExecutorFactory
from taurus.test import insertTest

#Define a "_NOT_PASSED" object to mark a keyword arg which is not passed
# Note that we do not want to use None because one may want to pass None
class __NotPassedType(int):
    pass
_NOT_PASSED = __NotPassedType()


def macroTest(klass=None, helper_name=None, test_method_name=None,
              test_method_doc=None, **helper_kwargs):
    """This decorator is an specialization of :function::`taurus.test.insertTest`
    for macro testing. It inserts test methods from a helper method that may 
    accept arguments.

    macroTest provides a very economic API for creating new tests for a given
    macro based on a helper method.

    macroTest accepts the following arguments:

        - helper_name (str): the name of the helper method. macroTest will
                             insert a test method which calls the helper with
                             any  the helper_kwargs (see below).
        - test_method_name (str): Optional. Name of the test method to be used.
                                 If None given, one will be generated from the
                                 macro and helper names.
        - test_method_doc (str): The docstring for the inserted test method
                                 (this shows in the unit test output). If None
                                 given, a default one is generated which
                                 includes the input parameters and the helper
                                 name.
        - \*\*helper_kwargs: All remaining keyword arguments are passed to the
                             helper.

    `macroTest` assumes that the decorated class has a `macro_name` 
    class member.

    This decorator can be considered a "base" decorator. It is often used to
    create other decorators in which the helper method is pre-set. Some
    of them are already provided in this module:

    - :meth:`testRun` is equivalent to macroTest with helper_name='macro_runs'
    - :meth:`testStop` is equivalent to macroTest with helper_name='macro_stops'
    - :meth:`testFail` is equivalent to macroTest with helper_name='macro_fails'

    The advantage of using the decorators compared to writing the test methods
    directly is that the helper method can get keyword arguments and therefore
    avoid duplication of code for very similar tests (think, e.g. on writing
    similar tests for various sets of macro input parameters):

    Consider the following code written using the
    :meth:`RunMacroTestCase.macro_runs` helper::

        class FooTest(RunMacroTestCase, unittest.TestCase)
            macro_name = twice

            def test_foo_runs_with_input_2(self):
                '''test that twice(2) runs'''
                self.macro_runs(macro_params=['2'])

            def test_foo_runs_with_input_minus_1(self):
                '''test that twice(2) runs'''
                self.macro_runs(macro_params=['-1'])

    The equivalent code could be written as::

        @macroTest(helper_name='macro_runs', macro_params=['2'])
        @macroTest(helper_name='macro_runs', macro_params=['-1'])
        class FooTest(RunMacroTestCase, unittest.TestCase):
            macro_name = 'twice'

    Or, even better, using the specialized testRun decorator::

        @testRun(macro_params=['2'])
        @testRun(macro_params=['-1'])
        class FooTest(RunMacroTestCase, unittest.TestCase):
            macro_name = 'twice'
            
    .. seealso:: :function::`taurus.test.insertTest`

    """
    #recipe to allow decorating with and without arguments
    if klass is None:
        return functools.partial(macroTest, helper_name=helper_name,
                                 test_method_name=test_method_name,
                                 test_method_doc=test_method_doc,
                                 **helper_kwargs)
        
    return insertTest(klass=klass, helper_name=helper_name, 
                      test_method_name=test_method_name,
                      test_method_doc=test_method_doc, 
                      tested_name = klass.macro_name, 
                      **helper_kwargs)


#Definition of specializations of the macroTest decorator:
testRun = functools.partial(macroTest, helper_name='macro_runs')
testStop = functools.partial(macroTest, helper_name='macro_stops')
testFail = functools.partial(macroTest, helper_name='macro_fails')


class BaseMacroTestCase(object):

    """An abstract class for macro testing.
    BaseMacroTestCase will provide a `macro_executor` member which is an
    instance of BaseMacroExecutor and which can be used to run a macro.

    To use it, simply inherit from BaseMacroTestCase *and* unittest.TestCase
    and provide the following class members:

      - macro_name (string) name of the macro to be tested (mandatory)
      - door_name (string) name of the door where the macro will be executed.
                 This is optional. If not set,
                 `sardanacustomsettings.UNITTEST_DOOR_NAME` is used

    Then you may define test methods.
    """
    macro_name = None
    door_name = getattr(sardanacustomsettings, 'UNITTEST_DOOR_NAME')

    def setUp(self):
        """ A macro_executor instance must be created
        """
        if self.macro_name is None:
            msg = '%s does not define macro_name' % self.__class__.__name__
            raise NotImplementedError(msg)

        mefact = MacroExecutorFactory()
        self.macro_executor = mefact.getMacroExecutor(self.door_name)

    def tearDown(self):
        """The macro_executor instance must be removed
        """
        self.macro_executor.unregisterAll()
        self.macro_executor = None


class RunMacroTestCase(BaseMacroTestCase):

    """A base class for testing execution of arbitrary Sardana macros.
    See :class:`BaseMacroTestCase` for requirements.

    It provides the following helper methods:
      - :meth:`macro_runs`
      - :meth:`macro_fails`
    """

    def assertFinished(self, msg):
        """Asserts that macro has finished.
        """
        finishStates = [u'finish']
        state = self.macro_executor.getState()
        #TODO buffer is just for debugging, attach only the last state
        state_buffer = self.macro_executor.getStateBuffer()
        msg = msg + '; State history=%s' % state_buffer
        self.assertIn(state, finishStates, msg)

    def setUp(self):
        """Preconditions:
        - Those from :class:`BaseMacroTestCase`
        - the macro executor registers to all the log levels
        """
        BaseMacroTestCase.setUp(self)
        self.macro_executor.registerAll()

    def macro_runs(self, macro_params=None, wait_timeout=float("inf"),
                   data=_NOT_PASSED):
        """A helper method to create tests that check if the macro can be
        successfully executed for the given input parameters. It may also
        optionally perform checks on the outputs from the execution.

        :param macro_params: (seq<str>): parameters for running the macro.
                             If passed, they must be given as a sequence of
                             their string representations.
        :param wait_timeout: (float) maximum allowed time (in s) for the macro
                             to finish. By default infinite timeout is used.
        :param data: (obj) Optional. If passed, the macro data after the
                     execution is tested to be equal to this.
        """
        self.macro_executor.run(macro_name=self.macro_name,
                                macro_params=macro_params,
                                sync=True, timeout=wait_timeout)
        self.assertFinished('Macro %s did not finish' % self.macro_name)

        #check if the data of the macro is the expected one
        if data is not _NOT_PASSED:
            actual_data = self.macro_executor.getData()
            msg = 'Macro data does not match expected data:\n' + \
                  'obtained=%s\nexpected=%s' % (actual_data, data)
            self.assertEqual(actual_data, data, msg)

        #TODO: implement generic asserts for macro result and macro output, etc
        #      in a similar way to what is done for macro data

    def macro_fails(self, macro_params=None, wait_timeout=float("inf"),
                    exception=None):
        """Check that the macro fails to run for the given input parameters

        :param macro_params: (seq<str>) input parameters for the macro
        :param wait_timeout: maximum allowed time for the macro to fail. By
                             default infinite timeout is used.
        :param exception: (str or Exception) if given, an additional check of
                        the type of the exception is done.
                        (IMPORTANT: this is just a comparison of str
                        representations of exception objects)
        """
        self.macro_executor.run(macro_name=self.macro_name,
                                macro_params=macro_params,
                                sync=True, timeout=wait_timeout)
        state = self.macro_executor.getState()
        actual_exc_str = self.macro_executor.getExceptionStr()
        msg = 'Post-execution state should be "exception" (got "%s")' % state
        self.assertEqual(state, 'exception', msg)

        if exception is not None:
            msg = 'Raised exception does not match expected exception:\n' + \
                  'raised=%s\nexpected=%s' % (actual_exc_str, exception)
            self.assertEqual(actual_exc_str, str(exception), msg)


class RunStopMacroTestCase(RunMacroTestCase):

    """This is an extension of :class:`RunMacroTestCase` to include helpers for
    testing the abort process of a macro. Useful for Runnable and Stopable
    macros.

    It provides the :meth:`macro_stops` helper
    """

    def assertStopped(self, msg):
        """Asserts that macro was stopped
        """
        stoppedStates = [u'stop']
        state = self.macro_executor.getState()
        #TODO buffer is just for debugging, attach only the last state
        state_buffer = self.macro_executor.getStateBuffer()
        msg = msg + '; State buffer was %s' % state_buffer
        self.assertIn(state, stoppedStates, msg)

    def macro_stops(self, macro_params=None, stop_delay=0.1,
                    wait_timeout=float("inf")):
        """A helper method to create tests that check if the macro can be
        successfully stoped (a.k.a. aborted) after it has been launched.

        :param macro_params: (seq<str>): parameters for running the macro.
                             If passed, they must be given as a sequence of
                             their string representations.
        :param stop_delay: (float) Time (in s) to wait between launching the
                           macro and sending the stop command. default=0.1
        :param wait_timeout: (float) maximum allowed time (in s) for the macro
                             to finish. By default infinite timeout is used.
        """
        self.macro_executor.run(macro_name=self.macro_name,
                                macro_params=macro_params,
                                sync=False)

        if stop_delay is not None:
            time.sleep(stop_delay)
        self.macro_executor.stop()
        self.macro_executor.wait(timeout=wait_timeout)
        self.assertStopped('Macro %s did not stop' % self.macro_name)


if __name__ == '__main__':
    from taurus.external import unittest
    from sardana.macroserver.macros.test import SarDemoEnv

    _m1 = SarDemoEnv().getMotors()[0]

    #@testRun(macro_params=[_m1, '0', '100', '4', '.1'])
    @testRun(macro_params=[_m1, '1', '0', '2', '.1'])
    @testRun(macro_params=[_m1, '0', '1', '4', '.1'])
    class dummyAscanTest(RunStopMacroTestCase, unittest.TestCase):
        macro_name = 'ascan'

    @testRun(macro_params=['1'], data={'in': 1, 'out': 2})
    @testRun(macro_params=['5'])
    @testRun
    class dummyTwiceTest(RunStopMacroTestCase, unittest.TestCase):
        macro_name = 'twice'

    @testFail
    @testFail(exception=Exception)
    class dummyRaiseException(RunStopMacroTestCase, unittest.TestCase):
        macro_name = 'raise_exception'

    suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        dummyRaiseException)
    unittest.TextTestRunner(descriptions=True, verbosity=2).run(suite)



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This package contains a collection of taurus widgets designed to connect
to sardana"""

__docformat__ = 'restructuredtext'

from taurus.external.qt import Qt
from taurus.qt.qtgui.resource import getIcon, getThemeIcon


class CommandLineHistory(list):
    pass

class CommandLine(Qt.QComboBox):

    def __init__(self, qt_parent=None, designMode=False):
        Qt.QComboBox.__init__(self, qt_parent)
        self.setEditable(True)
        self.setFrame(False)

class TaurusCommandLineWidget(Qt.QWidget):

    def __init__(self, qt_parent=None, designMode=False):
        Qt.QWidget.__init__(self, qt_parent)

        self._history = CommandLineHistory()

        l = Qt.QHBoxLayout()
        self.setLayout(l)
        l.setContentsMargins(0, 0, 0, 0)
        l.setSpacing(0)

        self._detailsButton = Qt.QToolButton()
        self._detailsButton.setText("...")

        self._cmdLine = Qt.QComboBox()
        self._cmdLine.setEditable(True)

        self._applyButton = Qt.QToolButton()
        self._applyButton.setIcon(getIcon(":/actions/media_playback_start.svg"))

        self._stopButton = Qt.QToolButton()
        self._stopButton.setIcon(getIcon(":/actions/media_playback_stop.svg"))

        self._clearButton = Qt.QToolButton()
        self._clearButton.setIcon(getThemeIcon("edit-clear"))

        l.addWidget(self._detailsButton, 0)
        l.addWidget(self._cmdLine, 1)
        l.addWidget(self._applyButton, 0)
        l.addWidget(self._stopButton, 0)
        l.addWidget(self._clearButton, 0)



    def run(self):
        pass

def demo():
    pass

def main():
    import sys
    import taurus.qt.qtgui.application
    Application = taurus.qt.qtgui.application.TaurusApplication

    app = Application.instance()
    owns_app = app is None

    if owns_app:
        import taurus.core.util.argparse
        parser = taurus.core.util.argparse.get_taurus_parser()
        app = Application(sys.argv, cmd_line_parser=parser,
                          app_name="Taurus command line demo", app_version="1.0",
                          org_domain="Taurus", org_name="Tango community")

    w = TaurusCommandLineWidget()
    w.show()

    if owns_app:
        sys.exit(app.exec_())
    else:
        return w

if __name__ == '__main__':
    main()



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""
macrodescriptionviewer.py: 
"""

import taurus.core
from taurus.external.qt import Qt
from taurus.qt.qtgui.base import TaurusBaseWidget


class TaurusMacroDescriptionViewer(Qt.QTextEdit, TaurusBaseWidget):
    __pyqtSignals__ = ("modelChanged(const QString &)",)

    def __init__(self, parent=None, designMode=False):
        name = "TaurusMacroDescriptionView"
        self.call__init__wo_kw(Qt.QTextEdit, parent)
        self.call__init__(TaurusBaseWidget, name)
        self.setReadOnly(True)
        self.setFont(Qt.QFont("Courier", 9))

    def defineStyle(self):
        """ Defines the initial style for the widget """
        self.updateStyle()

    def getModelClass(self):
        return taurus.core.taurusdevice.TaurusDevice

    def updateStyle(self):
        self.update()

    def onMacroNameChanged(self, macroName):
        """Can be connected to an event emitted after macro name was changed.
           As an argument receives macroName and ask BaseMacroServer object 
           about already prepared and stored in MacroInfoObj object macro description"""
        macroServer = self.getModelObj()

        if macroServer is None or macroName is None or macroName == "":
            self.setText("")
            return

        self.setText(str(macroServer.getMacroInfoObj(macroName).doc))

    def getFormatedToolTip(self, cache=True):
        """This method was overridden to get rid of the default tooltip of TaurusWidget"""
        return ""


    model = Qt.pyqtProperty("QString",
                                TaurusBaseWidget.getModel,
                                TaurusBaseWidget.setModel,
                                TaurusBaseWidget.resetModel)

    useParentModel = Qt.pyqtProperty("bool",
                                         TaurusBaseWidget.getUseParentModel,
                                         TaurusBaseWidget.setUseParentModel,
                                         TaurusBaseWidget.resetUseParentModel)


def test():
    import sys
    from sardana.taurus.core.tango.sardana.macroserver import registerExtensions
    registerExtensions()
    app = Qt.QApplication(sys.argv)
    taurusMacroDescriptionView = TaurusMacroDescriptionViewer(designMode=1)
    if len(sys.argv) != 2:
        taurusMacroDescriptionView.setModel("macroserver/zreszela/1")
    else:
        taurusMacroDescriptionView.setModel(sys.argv[1])
    taurusMacroDescriptionView.onMacroChanged("mv")
    taurusMacroDescriptionView.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    test()





#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

""" """

__all__ = ["IORegister", "IORegisterClass"]

__docformat__ = 'restructuredtext'

import sys
import time

from PyTango import DevFailed, Except
from PyTango import DevVoid, DevLong
from PyTango import DevState, AttrQuality
from PyTango import READ_WRITE, SCALAR

from taurus.core.util.log import DebugIt

from sardana import State, SardanaServer
from sardana.sardanaattribute import SardanaAttribute
from sardana.pool.poolexception import PoolException
from sardana.sardanautils import str_to_value

from sardana.tango.core.util import exception_str, throw_sardana_exception
from sardana.tango.pool.PoolDevice import PoolElementDevice, \
    PoolElementDeviceClass


class IORegister(PoolElementDevice):

    def __init__(self, dclass, name):
        self.in_write_value = False
        PoolElementDevice.__init__(self, dclass, name)

    def init(self, name):
        PoolElementDevice.init(self, name)

    def get_ior(self):
        return self.element

    def set_ior(self, ior):
        self.element = ior

    ior = property(get_ior, set_ior)

    def set_write_value_to_db(self):
        value_attr = self.ior.get_value_attribute()
        if value_attr.has_write_value():
            data = dict(Value=dict(__value=value_attr.w_value, __value_ts=value_attr.w_timestamp))
            db = self.get_database()
            db.put_device_attribute_property(self.get_name(), data)

    def get_write_value_from_db(self):
        name = 'Value'
        db = self.get_database()
        val_props = db.get_device_attribute_property(self.get_name(), name)[name]
        w_val = val_props["__value"][0]

        _, _, attr_info = self.get_dynamic_attributes()[0][name]
        w_val = str_to_value(w_val, attr_info.dtype, attr_info.dformat)

        w_val, w_ts = int(val_props["__value"][0]), None
        if "__value_ts" in val_props:
            w_ts = float(val_props["__value_ts"][0])
        return w_val, w_ts

    @DebugIt()
    def delete_device(self):
        PoolElementDevice.delete_device(self)
        ior = self.ior
        if ior is not None:
            ior.remove_listener(self.on_ior_changed)

    @DebugIt()
    def init_device(self):
        PoolElementDevice.init_device(self)

        ior = self.ior
        if ior is None:
            full_name = self.get_full_name()
            name = self.alias or full_name
            self.ior = ior = \
                self.pool.create_element(type="IORegister", name=name,
                    full_name=full_name, id=self.Id, axis=self.Axis,
                    ctrl_id=self.Ctrl_id)
            if self.instrument is not None:
                ior.set_instrument(self.instrument)
        ior.add_listener(self.on_ior_changed)

        ## force a state read to initialize the state attribute
        #state = ior.get_state(cache=False)
        self.set_state(DevState.ON)

    def on_ior_changed(self, event_source, event_type, event_value):
        try:
            self._on_ior_changed(event_source, event_type, event_value)
        except not DevFailed:
            msg = 'Error occurred "on_ior_changed(%s.%s): %s"'
            exc_info = sys.exc_info()
            self.error(msg, self.ior.name, event_type.name,
                       exception_str(*exc_info[:2]))
            self.debug("Details", exc_info=exc_info)

    def _on_ior_changed(self, event_source, event_type, event_value):
        # during server startup and shutdown avoid processing element
        # creation events
        if SardanaServer.server_state != State.Running:
            return

        timestamp = time.time()
        name = event_type.name.lower()

        if name == "w_value" and not self.in_write_value:
            self.debug("Storing value set point: %s", self.ior.value.w_value)
            self.set_write_value_to_db()
            return

        multi_attr = self.get_device_attr()
        try:
            attr = multi_attr.get_attr_by_name(name)
        except DevFailed:
            return

        quality = AttrQuality.ATTR_VALID
        priority = event_type.priority
        value, w_value, error = None, None, None

        if name == "state":
            value = self.calculate_tango_state(event_value)
        elif name == "status":
            value = self.calculate_tango_status(event_value)
        else:
            if isinstance(event_value, SardanaAttribute):
                if event_value.error:
                    error = Except.to_dev_failed(*event_value.exc_info)
                else:
                    value = event_value.value
                timestamp = event_value.timestamp
            state = self.ior.get_state(propagate=0)

            if name == "value":
                w_value = event_source.get_value_attribute().w_value
                if state == State.Moving:
                    quality = AttrQuality.ATTR_CHANGING
        self.set_attribute(attr, value=value, w_value=w_value,
                           timestamp=timestamp, quality=quality,
                           priority=priority, error=error, synch=False)

    def always_executed_hook(self):
        #state = to_tango_state(self.ior.get_state(cache=False))
        pass

    def read_attr_hardware(self, data):
        pass

    def initialize_dynamic_attributes(self):
        attrs = PoolElementDevice.initialize_dynamic_attributes(self)

        detect_evts = "value",
        non_detect_evts = ()

        for attr_name in detect_evts:
            if attrs.has_key(attr_name):
                self.set_change_event(attr_name, True, True)
        for attr_name in non_detect_evts:
            if attrs.has_key(attr_name):
                self.set_change_event(attr_name, True, False)
        return

    def read_Value(self, attr):
        value = self.ior.get_value(cache=False)
        if value.error:
            Except.throw_python_exception(*value.exc_info)
        self.set_attribute(attr, value=value.value, w_value=value.w_value,
                           priority=0, timestamp=value.timestamp)

    def write_Value(self, attr):
        self.in_write_value = True
        value = attr.get_write_value()
        try:
            self.ior.set_value(value)
            # manually store write value in the database
            self.set_write_value_to_db()
        except PoolException as pe:
            throw_sardana_exception(pe)
        finally:
            self.in_write_value = False

    def is_Value_allowed(self, req_type):
        if self.get_state() in [DevState.FAULT, DevState.UNKNOWN]:
            return False
        return True

    def Start(self):
        self.ior.start_acquisition()


class IORegisterClass(PoolElementDeviceClass):

    #    Class Properties
    class_property_list = {
    }

    #    Device Properties
    device_property_list = {
    }
    device_property_list.update(PoolElementDeviceClass.device_property_list)

    #    Command definitions
    cmd_list = {
        'Start' :   [ [DevVoid, ""], [DevVoid, ""] ],
    }
    cmd_list.update(PoolElementDeviceClass.cmd_list)

    #    Attribute definitions
    attr_list = {}
    attr_list.update(PoolElementDeviceClass.attr_list)

    standard_attr_list = {
        'Value'     : [ [ DevLong, SCALAR, READ_WRITE ],
                        { 'Memorized'     : "true_without_hard_applied", }, ],
    }
    standard_attr_list.update(PoolElementDeviceClass.standard_attr_list)

    def _get_class_properties(self):
        ret = PoolElementDeviceClass._get_class_properties(self)
        ret['Description'] = "IORegister device class"
        ret['InheritedFrom'].insert(0, 'PoolElementDevice')
        return ret



#!/usr/bin/env python

##############################################################################
##
## This file is part of Sardana
##
## http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
## 
## Sardana is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
## 
## Sardana is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
## 
## You should have received a copy of the GNU Lesser General Public License
## along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Pool libray. It defines the base classes
for a pool container element"""

__all__ = ["PoolContainer"]

__docformat__ = 'restructuredtext'

from sardana.sardanacontainer import SardanaContainer


class PoolContainer(SardanaContainer):
    """A container class for pool elements"""
    
    def get_controller_class(self, **kwargs):
        eid = kwargs.get("id")
        if eid is not None:
            return self.get_controller_class_by_id(eid, **kwargs)
        
        name = kwargs.pop("name")
        self.get_controller_class_by_name(name, **kwargs)
    
    def get_controller_class_by_id(self, eid, **kwargs):
        raise NotImplementedError
    
    def get_controller_class_by_name(self, name, **kwargs):
        raise NotImplementedError